import numpy as np
from xarray import open_dataset, Dataset

from foxes.core import FarmController
import foxes.constants as FC
import foxes.variables as FV


class OpFlagController(FarmController):
    """
    A basic controller with a flag for
    turbine operation at each state.

    Parameters
    ----------
    non_op_values: dict
        The non-operational values for variables,
        keys: variable str, values: float
    var2ncvar: dict
        The mapping of variable names to NetCDF variable names,
        only needed if data_source is a path to a NetCDF file

    :group: models.farm_controllers

    """

    def __init__(
        self,
        data_source,
        non_op_values=None,
        var2ncvar={},
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        data_source: numpy.ndarray or str
            The operating flag data, shape: (n_states, n_turbines),
            or path to a NetCDF file
        non_op_values: dict, optional
            The non-operational values for variables,
            keys: variable str, values: float
        var2ncvar: dict
            The mapping of variable names to NetCDF variable names,
            only needed if data_source is a path to a NetCDF file
        kwargs: dict, optional
            Additional keyword arguments for the
            base class constructor

        """
        super().__init__(**kwargs)
        self.data_source = data_source
        self.var2ncvar = var2ncvar

        self.non_op_values = {
            FV.P: 0.0,
            FV.CT: 0.0,
        }
        if non_op_values is not None:
            self.non_op_values.update(non_op_values)

        self._op_flags = None

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        vrs = set(super().output_farm_vars(algo))
        vrs.update([FV.OPERATING])
        return list(vrs)

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """

        idata = super().load_data(algo, verbosity)

        if isinstance(self.data_source, np.ndarray):
            self._op_flags = self.data_source

        elif isinstance(self.data_source, Dataset):
            cop = self.var2ncvar.get(FV.OPERATING, FV.OPERATING)
            self._op_flags = self.data_source[cop].to_numpy()

        else:
            if verbosity > 0:
                print(f"OpFlagController: Reading data from {self.data_source}")
            ds = open_dataset(self.data_source)
            cop = self.var2ncvar.get(FV.OPERATING, FV.OPERATING)
            self._op_flags = ds[cop].to_numpy()
            del ds

        assert self._op_flags.shape == (algo.n_states, algo.n_turbines), (
            f"OpFlagController data shape {self._op_flags.shape} does not match "
            f"(n_states, n_turbines)=({algo.n_states}, {algo.n_turbines})"
        )
        op_flags = self._op_flags.astype(bool)

        off = np.where(~op_flags)
        tmsels = idata["data_vars"][FC.TMODEL_SELS][1]
        tmsels[off[0], off[1], :] = False
        self._tmall = [np.all(t) for t in tmsels]

        idata["data_vars"][FC.TMODEL_SELS] = (
            (FC.STATE, FC.TURBINE, FC.TMODELS),
            tmsels,
        )
        idata["data_vars"][FV.OPERATING] = (
            (FC.STATE, FC.TURBINE),
            op_flags,
        )

        return idata

    def calculate(self, algo, mdata, fdata, pre_rotor, downwind_index=None):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        pre_rotor: bool
            Flag for running pre-rotor or post-rotor
            models
        downwind_index: int, optional
            The index in the downwind order

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        self.ensure_output_vars(algo, fdata)

        # compute data for all operating turbines:
        op = mdata[FV.OPERATING].astype(bool)
        fdata[FV.OPERATING] = op
        results = super().calculate(algo, mdata, fdata, pre_rotor, downwind_index)
        results[FV.OPERATING] = fdata[FV.OPERATING]

        # set non-operating values:
        if downwind_index is None:
            off = np.where(~op)
            for v in self.output_farm_vars(algo):
                if v != FV.OPERATING:
                    fdata[v][off[0], off[1]] = self.non_op_values.get(v, np.nan)
        else:
            off = np.where(~op[:, downwind_index])
            for v in self.output_farm_vars(algo):
                if v != FV.OPERATING:
                    fdata[v][off[0], downwind_index] = self.non_op_values.get(v, np.nan)

        return results



import numpy as np

from foxes.utils import Dict
import foxes.variables as FV
import foxes.constants as FC

from .read_fields import foxes2wio


def _read_turbine_outputs(wio_outs, olist, algo, states_isel, verbosity):
    """Reads the turbine outputs request"""
    if "turbine_outputs" in wio_outs and wio_outs["turbine_outputs"].get_item(
        "report", True
    ):
        turbine_outputs = wio_outs["turbine_outputs"]
        turbine_nc_filename = turbine_outputs.pop_item(
            "turbine_nc_filename", "turbine_outputs.nc"
        )
        output_variables = turbine_outputs["output_variables"]
        if verbosity > 2:
            print("      Reading turbine_outputs")
            print("        File name:", turbine_nc_filename)
            print("        output_variables:", output_variables)

        vmap = Dict(
            power=FV.P,
            rotor_effective_velocity=FV.REWS,
            _name="vmap",
        )
        ivmap = {d: k for k, d in vmap.items()}
        ivmap.update(
            {
                FC.STATE: "time",
                FC.TURBINE: "turbine",
            }
        )

        isel = None
        if states_isel is not None and len(states_isel):
            isel = {"time": states_isel}

        olist.append(
            Dict(
                output_type="StateTurbineTable",
                functions=[
                    dict(
                        function="get_dataset",
                        variables=[vmap[v] for v in output_variables],
                        name_map=ivmap,
                        to_file=turbine_nc_filename,
                        round={
                            vw: FV.get_default_digits(vf) for vw, vf in vmap.items()
                        },
                        isel=isel,
                        verbosity=verbosity,
                    )
                ],
                _name=f"outputs.{len(olist)}.StateTurbineTable",
            )
        )


def _read_flow_field(wio_outs, olist, algo, states_isel, verbosity):
    """Reads the flow field request"""
    if "flow_field" in wio_outs and wio_outs["flow_field"].get_item("report", True):
        flow_field = wio_outs["flow_field"]
        flow_nc_filename = flow_field.pop_item("flow_nc_filename", "flow_field.nc")
        output_variables = flow_field.pop_item("output_variables")

        z_planes = flow_field.pop_item("z_planes")
        z_sampling = z_planes["z_sampling"]
        xy_sampling = z_planes["xy_sampling"]

        if verbosity > 2:
            print("      Reading flow_field")
            print("        File name       :", flow_nc_filename)
            print("        output_variables:", output_variables)
            print("        states subset   :", states_isel)
            print("        z_sampling      :", z_sampling)
            print("        xy_sampling     :", xy_sampling)

        vmap = Dict(
            wind_speed=FV.WS,
            wind_direction=FV.WD,
            _name="vmap",
        )

        z_list = []
        if z_sampling == "plane_list":
            z_list = z_planes.pop_item("z_list")
        elif z_sampling == "hub_heights":
            z_list = np.unique(algo.farm.get_hub_heights(algo))
        elif z_sampling == "grid":
            zb = z_planes.pop_item("z_bounds")
            assert len(zb) == 2, f"Expecting two entries for z_bounds, got {zb}"
            zn = z_planes.pop_item("z_number")
            z_list = np.linspace(zb[0], zb[1], zn)
        elif isinstance(z_sampling, (int, float)):
            z_list = np.atleast_1d(z_sampling)
        else:
            raise NotImplementedError(
                f"z_sampling '{z_sampling}' is not supported. Choices: plane_list, hub_heights, grid."
            )
        z_list = np.asarray(z_list)
        if verbosity > 2:
            print("          z_list        :", z_list)

        if xy_sampling == "grid":
            xb = z_planes.pop_item("x_bounds")
            assert len(xb) == 2, f"Expecting two entries for x_bounds, got {xb}"
            yb = z_planes.pop_item("y_bounds")
            assert len(yb) == 2, f"Expecting two entries for y_bounds, got {yb}"
            if "dx" in z_planes or "dy" in z_planes:
                assert "dx" in z_planes and "dy" in z_planes, (
                    f"Expecting both 'dx' and 'dy' in z_planes, got {list(z_planes.keys())}"
                )
                dx = z_planes.pop_item("dx")
                dy = z_planes.pop_item("dy")
                nx = max(int((xb[1] - xb[0]) / dx), 1) + 1
                if (xb[1] - xb[0]) / (nx - 1) > dx:
                    nx += 1
                ny = max(int((yb[1] - yb[0]) / dy), 1) + 1
                if (yb[1] - yb[0]) / (ny - 1) > dy:
                    ny += 1
            elif "Nx" in z_planes or "Ny" in z_planes:
                assert "Nx" in z_planes and "Ny" in z_planes, (
                    f"Expecting both 'Nx' and 'Ny' in z_planes, got {list(z_planes.keys())}"
                )
                nx = z_planes.pop_item("Nx")
                ny = z_planes.pop_item("Ny")
            else:
                raise KeyError(
                    f"Expecting either 'dx' and 'dy' or 'Nx' and 'Ny' in z_planes, got {list(z_planes.keys())}"
                )
            z_list = np.asarray(z_list)
            if verbosity > 2:
                print("          x_bounds      :", xb)
                print("          y_bounds      :", yb)
                print("          nx, ny        :", (nx, ny))
                print("          true dx       :", (xb[1] - xb[0]) / (nx - 1))
                print("          true dy       :", (yb[1] - yb[0]) / (ny - 1))
            olist.append(
                Dict(
                    output_type="SlicesData",
                    verbosity_delta=3,
                    functions=[
                        dict(
                            function="get_states_data_xy",
                            z_list=z_list,
                            states_isel=states_isel,
                            xmin=xb[0],
                            xmax=xb[1],
                            ymin=yb[0],
                            ymax=yb[1],
                            n_img_points=(nx, ny),
                            variables=[vmap[v] for v in output_variables],
                            to_file=flow_nc_filename,
                            label_map=foxes2wio,
                            verbosity=verbosity,
                        )
                    ],
                    _name=f"outputs.output{len(olist)}.SliceData",
                )
            )
        else:
            raise NotImplementedError(
                f"xy_sampling '{xy_sampling}' is not supported. Choices: 'grid'."
            )


def read_outputs(wio_outs, idict, algo, verbosity=1):
    """
    Reads the windio outputs

    Parameters
    ----------
    wio_outs: foxes.utils.Dict
        The windio output data dict
    idict: foxes.utils.Dict
        The foxes input data dictionary
    algo: foxes.core.Algorithm
        The algorithm
    verbosity: int
        The verbosity level, 0=silent

    Returns
    -------
    odir: pathlib.Path
        The output directory

    :group: input.yaml.windio

    """
    odir = wio_outs.pop_item("output_folder", ".")
    olist = []
    if verbosity > 2:
        print("  Reading model_outputs_specification")
        print("    Output dir:", odir)
        print("    Contents  :", [k for k in wio_outs.keys()])

    # read subset:
    run_configuration = wio_outs.pop_item("run_configuration", {})
    states_isel = None
    if "times_run" in run_configuration:
        times_run = run_configuration.pop_item("times_run")
        if not times_run.get_item("all_occurences"):
            states_isel = times_run.get_item("subset")
    elif "wind_speeds_run" in run_configuration:
        wind_speeds_run = run_configuration.get_item("wind_speeds_run")
        directions_run = run_configuration.get_item("directions_run")
        if not wind_speeds_run.get_item("all_values"):
            raise NotImplementedError(
                f"Wind speed and direction subsets are not yet supported, got {wind_speeds_run.name} {wind_speeds_run}"
            )
        if not directions_run.get_item("all_values"):
            raise NotImplementedError(
                f"Wind speed and direction subsets are not yet supported, got {directions_run.name} {directions_run}"
            )

    # read turbine_outputs:
    _read_turbine_outputs(wio_outs, olist, algo, states_isel, verbosity)

    # read flow field:
    _read_flow_field(wio_outs, olist, algo, states_isel, verbosity)

    if len(olist):
        idict["outputs"] = olist

    return odir



"""
Rotor models.
"""

from .centre import CentreRotor as CentreRotor
from .grid import GridRotor as GridRotor
from .levels import LevelRotor as LevelRotor
from .direct_infusion import DirectMDataInfusion as DirectMDataInfusion



import numpy as np
import pandas as pd

from foxes.core import TurbineType, FData
from foxes.data import parse_Pct_file_name
from foxes.models.turbine_models import LookupTable
import foxes.variables as FV


class FromLookupTable(TurbineType):
    """
    Calculate power and ct by interpolating
    by using a lookup-table

    Attributes
    ----------
    source: str or pandas.DataFrame
        The file path, static name, or data
    rho: float
        The air density for which the data is valid
        or None for no correction
    WSCT: str
        The wind speed variable for ct lookup
    WSP: str
        The wind speed variable for power lookup
    rpars: dict, optional
        Parameters for pandas file reading

    :group: models.turbine_types

    """

    def __init__(
        self,
        data_source,
        input_vars,
        varmap={},
        lookup_pars={},
        rho=None,
        p_ct=1.0,
        p_P=1.88,
        var_ws_ct=FV.REWS2,
        var_ws_P=FV.REWS3,
        pd_file_read_pars={},
        interpn_args={},
        **parameters,
    ):
        """
        Constructor.

        Parameters
        ----------
        data_source: str or pandas.DataFrame
            The file path, static name, or data
        input_vars: list of str
            The foxes input variables
        varmap: dict
            Mapping from foxes variable names
            to column names in the data_source
        lookup_pars: dict
            Additional parameters for the LookupTable model
        rho: float, optional
            The air density for which the data is valid
            or None for no correction
        p_ct: float
            The exponent for yaw dependency of ct
        p_P: float
            The exponent for yaw dependency of P
        var_ws_ct: str
            The wind speed variable for ct lookup
        var_ws_P: str
            The wind speed variable for power lookup
        pd_file_read_pars: dict
            Parameters for pandas file reading
        interpn_args: dict
            Parameters for scipy intern or interp1d
        parameters: dict, optional
            Additional parameters for TurbineType class

        """
        if not isinstance(data_source, pd.DataFrame):
            pars = parse_Pct_file_name(data_source)
            pars.update(parameters)
        else:
            pars = parameters

        super().__init__(**pars)

        self.source = data_source
        self.rho = rho
        self.p_ct = p_ct
        self.p_P = p_P
        self.WSCT = var_ws_ct
        self.WSP = var_ws_P
        self.rpars = pd_file_read_pars

        if FV.REWS not in input_vars or len(
            set(input_vars).intersection([FV.WS, FV.REWS2, FV.REWS3])
        ):
            raise KeyError(
                f"Turbine type '{self.name}': Expecting '{FV.REWS}' as wind speed variable in inputv_vars, got {input_vars}"
            )

        iargs = dict(bounds_error=False, fill_value=0)
        iargs.update(interpn_args)
        self._lookup = LookupTable(
            data_source=data_source,
            input_vars=input_vars,
            output_vars=[FV.P, FV.CT],
            varmap=varmap,
            interpn_args=iargs,
            **lookup_pars,
        )

    def __repr__(self):
        a = f"D={self.D}, H={self.H}, P_nominal={self.P_nominal}, P_unit={self.P_unit}, rho={self.rho}"
        a += f", var_ws_ct={self.WSCT}, var_ws_P={self.WSP}"
        return f"{type(self).__name__}({a})"

    def needs_rews2(self):
        """
        Returns flag for requiring REWS2 variable

        Returns
        -------
        flag: bool
            True if REWS2 is required

        """
        return self.WSCT == FV.REWS2 or self.WSP == FV.REWS2

    def needs_rews3(self):
        """
        Returns flag for requiring REWS3 variable

        Returns
        -------
        flag: bool
            True if REWS3 is required

        """
        return self.WSCT == FV.REWS3 or self.WSP == FV.REWS3

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            All sub models

        """
        return [self._lookup]

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return [FV.P, FV.CT]

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        super().initialize(algo, verbosity, force)
        if self.P_nominal is None:
            col_P = self._lookup.varmap.get(FV.P, FV.P)
            self.P_nominal = np.max(self._lookup._data[col_P].to_numpy())

    def modify_cutin(
        self,
        modify_ct,
        modify_P,
        steps=20,
        iterations=100,
        a=0.55,
        b=0.55,
    ):
        """
        Modify the data such that a discontinuity
        at cutin wind speed is avoided

        Parameters
        ----------
        variable: str
            The target variable
        modify_ct: bool
            Flag for modification of the ct curve
        modify_P: bool
            Flag for modification of the power curve
        steps: int
            The number of wind speed steps between 0 and
            the cutin wind speed
        iterations: int
            The number of iterations
        a: float
            Coefficient for iterative mixing
        b: float
            Coefficient for iterative mixing

        """
        if modify_ct or modify_P:
            raise NotImplementedError

        else:
            super().modify_cutin(modify_ct, modify_P)

    def calculate(self, algo, mdata, fdata, st_sel):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        st_sel: numpy.ndarray of bool
            The state-turbine selection,
            shape: (n_states, n_turbines)

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        # prepare data for lookup:
        input_vars = self._lookup.input_vars
        fdata_lookup = FData(
            data={v: fdata[v] for v in input_vars},
            dims={v: fdata.dims[v] for v in input_vars},
            loop_dims=fdata.loop_dims,
        )
        for v in self.output_farm_vars(algo):
            fdata_lookup.add(v, fdata[v], fdata.dims[v])

        rews2 = None
        rews3 = None
        if self.WSP != FV.REWS or self.WSCT != FV.REWS:
            rews2 = fdata[self.WSCT].copy()
            rews3 = fdata[self.WSP].copy()

        # apply air density correction:
        if self.rho is not None:
            if rews2 is None:
                rews2 = fdata[self.WSCT].copy()
                rews3 = fdata[self.WSP].copy()

            # correct wind speed by air density, such
            # that in the partial load region the
            # correct value is reconstructed:
            rho = fdata[FV.RHO][st_sel]
            rews3[st_sel] *= (self.rho / rho) ** (1.0 / 3.0)
            del rho

        # in yawed case, calc yaw corrected wind speed:
        if FV.YAWM in fdata and (self.p_P is not None or self.p_ct is not None):
            if rews2 is None:
                rews2 = fdata[self.WSCT].copy()
                rews3 = fdata[self.WSP].copy()

            # calculate corrected wind speed wsc,
            # gives ws**3 * cos**p_P in partial load region
            # and smoothly deals with full load region:
            yawm = fdata[FV.YAWM][st_sel]
            if np.any(np.isnan(yawm)):
                raise ValueError(
                    f"{self.name}: Found NaN values for variable '{FV.YAWM}'. Maybe change order in turbine_models?"
                )
            cosm = np.cos(yawm / 180 * np.pi)
            if self.p_ct is not None:
                rews2[st_sel] *= (cosm**self.p_ct) ** 0.5
            if self.p_P is not None:
                rews3[st_sel] *= (cosm**self.p_P) ** (1.0 / 3.0)
            del yawm, cosm

        # run lookup:
        if rews2 is None:
            out = self._lookup.calculate(algo, mdata, fdata_lookup, st_sel)
        else:
            fdata_lookup[FV.REWS] = rews2
            ct = self._lookup.calculate(algo, mdata, fdata_lookup, st_sel)[FV.CT]
            fdata_lookup[FV.REWS] = rews3
            out = self._lookup.calculate(algo, mdata, fdata_lookup, st_sel)
            out[FV.CT] = ct

        return out



from .load import import_module


def print_mem(obj, min_csize=0, max_csize=None, pre_str="OBJECT SIZE"):
    """
    Prints the memory consumption of a model and its components

    Parmeters
    ---------
    obj: object
        The object to be analyzed
    min_csize: int
        The minimal size of a component for being shown
    max_csize: int, optional
        The maximal allowed size of a component
    pre_str: str
        String to be printed before

    :group: utils

    """
    objsize = import_module("objsize")
    n = obj.name if hasattr(obj, "name") else ""
    print(pre_str, type(obj).__name__, n, objsize.get_deep_size(obj))
    for k in dir(obj):
        o = None
        try:
            if (
                hasattr(obj, k)
                and not callable(getattr(obj, k))
                and (len(k) < 3 or k[:2] != "__")
            ):
                o = getattr(obj, k)
        except ValueError:
            pass

        if o is not None:
            s = objsize.get_deep_size(getattr(obj, k))
            if s >= min_csize:
                print("   ", k, s)
                if max_csize is not None and s > max_csize:
                    raise ValueError(f"Component {k} exceeds maximal size {max_csize}")



import numpy as np

from foxes.config import config
from foxes.core import TurbineInductionModel
import foxes.variables as FV
import foxes.constants as FC


class RankineHalfBody(TurbineInductionModel):
    """
    The Rankine half body induction wake model

    Notes
    -----
    Reference:
    B Gribben and G Hawkes
    "A potential flow model for wind turbine induction and wind farm blockage"
    Techincal Paper, Frazer-Nash Consultancy, 2019
    https://www.fnc.co.uk/media/o5eosxas/a-potential-flow-model-for-wind-turbine-induction-and-wind-farm-blockage.pdf

    Attributes
    ----------
    induction: foxes.core.AxialInductionModel or str
        The induction model

    :group: models.wake_models.induction

    """

    def __init__(self, superposition="vector", induction="Madsen"):
        """
        Constructor.

        Parameters
        ----------
        superposition: str
            The wind speed deficit superposition.
        induction: foxes.core.AxialInductionModel or str
            The induction model

        """
        super().__init__(wind_superposition=superposition, other_superpositions={})
        self.induction = induction

        self._has_uv = True

    def __repr__(self):
        iname = (
            self.induction if isinstance(self.induction, str) else self.induction.name
        )
        return f"{type(self).__name__}({self.wind_superposition}, induction={iname})"

    @property
    def affects_ws(self):
        """
        Flag for wind speed wake models

        Returns
        -------
        dws: bool
            If True, this model affects wind speed

        """
        return True

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            All sub models

        """
        return super().sub_models() + [self.induction]

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        if isinstance(self.induction, str):
            self.induction = algo.mbook.axial_induction[self.induction]
        super().initialize(algo, verbosity, force)

    def new_wake_deltas(self, algo, mdata, fdata, tdata):
        """
        Creates new empty wake delta arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        wake_deltas: dict
            Key: variable name, value: The zero filled
            wake deltas, shape: (n_states, n_targets, n_tpoints, ...)

        """
        duv = np.zeros(
            (tdata.n_states, tdata.n_targets, tdata.n_tpoints, 2),
            dtype=config.dtype_double,
        )
        return {FV.UV: duv}

    def contribute(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        wake_coos,
        wake_deltas,
    ):
        """
        Modifies wake deltas at target points by
        contributions from the specified wake source turbines.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints, ...)

        """
        # get ct:
        ct = self.get_data(
            FV.CT,
            FC.STATE_TARGET_TPOINT,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=False,
        )

        # get ws:
        ws = self.get_data(
            FV.REWS,
            FC.STATE_TARGET_TPOINT,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=False,
        )

        # get D
        D = self.get_data(
            FV.D,
            FC.STATE_TARGET_TPOINT,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=True,
        )

        # calc m (page 7, skipping pi everywhere)
        m = 2 * ws * self.induction.ct2a(ct) * (D / 2) ** 2

        # get r and theta
        x = np.round(wake_coos[..., 0], 12)
        r = np.linalg.norm(wake_coos[..., 1:], axis=-1)
        r_sph = np.sqrt(r**2 + x**2)
        theta = np.arctan2(r, x)

        # define rankine half body shape (page 3)
        RHB_shape = (
            np.cos(theta) - (2 / (m + 1e-15)) * ws * (r_sph * np.sin(theta)) ** 2
        )

        # stagnation point condition
        xs = -np.sqrt(m / (4 * ws + 1e-15))

        # set values out of body shape
        st_sel = (ct > 1e-8) & ((RHB_shape < -1) | (x < xs))
        if np.any(st_sel):
            # apply selection
            xyz = wake_coos[st_sel]

            # calc velocity components
            vel_factor = m[st_sel] / (4 * np.linalg.norm(xyz, axis=-1) ** 3)
            wake_deltas[FV.UV][st_sel] += vel_factor[:, None] * xyz[:, :2]

        # set values inside body shape
        st_sel = (ct > 1e-8) & (RHB_shape >= -1) & (x >= xs) & (x <= 0)
        if np.any(st_sel):
            # apply selection
            xyz = np.zeros_like(wake_coos[st_sel])
            xyz[:, 0] = xs[st_sel]

            # calc velocity components
            vel_factor = m[st_sel] / (4 * np.linalg.norm(xyz, axis=-1) ** 3)
            wake_deltas[FV.UV][st_sel, 0] += vel_factor * xyz[:, 0]



from yaml import safe_load
from pathlib import Path


class Dict(dict):
    """
    A slightly enhanced dictionary.

    :group: utils

    """

    def __init__(self, *args, _name=None, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        *args: tuple, optional
            Arguments passed to `dict`
        _name: str, optional
            The dictionary name
        **kwargs: dict, optional
            Arguments passed to `dict`

        """
        super().__init__()
        self._name = _name if _name is not None else type(self).__name__
        self.update(*args, **kwargs)

    @property
    def name(self):
        """
        The dictionary name

        Returns
        -------
        name: str
            The dictionary name

        """
        return self._name

    def get_item(self, key, *deflt, prnt=True):
        """
        Gets an item, prints readable error if not found

        Parameters
        ----------
        key: immutable object
            The key
        deflt: tuple, optional
            Tuple of length 1, containing the default
        prnt: bool
            Flag for message printing

        Returns
        -------
        data: object
            The data

        """
        try:
            if len(deflt):
                assert len(deflt) == 1, (
                    f"Expecting a single default entry, got {len(deflt)}"
                )
                data = self.get(key, deflt[0])
            else:
                data = self[key]
        except KeyError as e:
            if prnt:
                print(f"\n{self.name}: Cannot find key '{key}'.\n")
                print("Known keys:")
                for k in self.keys():
                    print("   ", k)
                print()
            raise e

        if isinstance(data, dict) and not isinstance(data, Dict):
            data = Dict(data, _name=f"{self.name}.{key}")

        return data

    def pop_item(self, key, *deflt, prnt=True):
        """
        Pops an item, prints readable error if not found

        Parameters
        ----------
        key: immutable object
            The key
        deflt: tuple, optional
            Tuple of length 1, containing the default
        prnt: bool
            Flag for message printing

        Returns
        -------
        data: object
            The data

        """
        data = self.get_item(key, *deflt, prnt=prnt)
        if key in self:
            del self[key]
        return data

    def __setitem__(self, key, value):
        if isinstance(value, list):
            out = []
            for i, x in enumerate(value):
                if isinstance(x, dict) and not isinstance(x, Dict):
                    nme = f"{self.name}.{key}"
                    if len(value) > 1:
                        nme += f".{i}"
                    out.append(Dict(x, _name=nme))
                else:
                    out.append(x)
            value = out
        elif isinstance(value, dict) and not isinstance(value, Dict):
            out = Dict(_name=f"{self.name}.{key}")
            out.update(value)
            value = out

        super().__setitem__(key, value)

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            k = ", ".join(sorted([f"{s}" for s in self.keys()]))
            e = f"{self.name}: Cannot find key '{key}'. Known keys: {k}"
            raise KeyError(e)

    def update(self, *args, **kwargs):
        """
        Update the dictionary with the key/value pairs from other, overwriting existing keys.
        """
        other = dict(*args, **kwargs)
        for k, v in other.items():
            self[k] = v

    @classmethod
    def from_yaml(self, yml_file, verbosity=1):
        """
        Reads a yaml file

        Parameters
        ----------
        yml_file: str
            Path to the yaml file
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        dct: Dict
            The data

        """

        def _print(*args, level=1, **kwargs):
            if verbosity >= level:
                print(*args, **kwargs)

        fpath = Path(yml_file)
        _print("Reading file", fpath)
        with open(fpath) as stream:
            data = safe_load(stream)
        if data is None:
            data = {}
        dct = Dict(data, _name=fpath.stem)
        _print(dct, level=2)

        return dct



import numpy as np

from foxes.config import config
from foxes.core import TurbineInductionModel
import foxes.variables as FV
import foxes.constants as FC


class VortexSheet(TurbineInductionModel):
    """
    The Vortex Sheet model implemented with a radial dependency

    Notes
    -----
    Reference:
    Medici, D., et al. "The upstream flow of a wind turbine: blockage effect." Wind Energy 14.5 (2011): 691-697.
    https://doi.org/10.1002/we.451

    Attributes
    ----------
    pre_rotor_only: bool
        Calculate only the pre-rotor region
    induction: foxes.core.AxialInductionModel or str
        The induction model

    :group: models.wake_models.induction

    """

    def __init__(
        self,
        superposition="ws_linear",
        induction="Madsen",
        pre_rotor_only=False,
    ):
        """
        Constructor.

        Parameters
        ----------
        superposition: str
            The wind speed superposition
        induction: foxes.core.AxialInductionModel or str
            The induction model
        pre_rotor_only: bool
            Calculate only the pre-rotor region

        """
        super().__init__(wind_superposition=superposition)
        self.induction = induction
        self.pre_rotor_only = pre_rotor_only

    def __repr__(self):
        iname = (
            self.induction if isinstance(self.induction, str) else self.induction.name
        )
        return f"{type(self).__name__}({self.wind_superposition}, induction={iname})"

    @property
    def affects_ws(self):
        """
        Flag for wind speed wake models

        Returns
        -------
        dws: bool
            If True, this model affects wind speed

        """
        return True

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            All sub models

        """
        return super().sub_models() + [self.induction]

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        if isinstance(self.induction, str):
            self.induction = algo.mbook.axial_induction[self.induction]
        super().initialize(algo, verbosity, force)

    def new_wake_deltas(self, algo, mdata, fdata, tdata):
        """
        Creates new empty wake delta arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        wake_deltas: dict
            Key: variable name, value: The zero filled
            wake deltas, shape: (n_states, n_targets, n_tpoints, ...)

        """
        if self.has_uv:
            duv = np.zeros(
                (tdata.n_states, tdata.n_targets, tdata.n_tpoints, 2),
                dtype=config.dtype_double,
            )
            return {FV.UV: duv}
        else:
            dws = np.zeros(
                (tdata.n_states, tdata.n_targets, tdata.n_tpoints),
                dtype=config.dtype_double,
            )
            return {FV.WS: dws}

    def contribute(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        wake_coos,
        wake_deltas,
    ):
        """
        Modifies wake deltas at target points by
        contributions from the specified wake source turbines.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data
        tdata: foxes.core.Data
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints, ...)

        """

        # get x, y and z. Rounding for safe x < 0 condition
        x = np.round(wake_coos[..., 0], 12)
        y = wake_coos[..., 1]
        z = wake_coos[..., 2]
        r = np.sqrt(y**2 + z**2)
        r_sph = np.sqrt(r**2 + x**2)

        # get ct
        ct = self.get_data(
            FV.CT,
            FC.STATE_TARGET_TPOINT,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            upcast=True,
            downwind_index=downwind_index,
        )

        # get D
        D = self.get_data(
            FV.D,
            FC.STATE_TARGET_TPOINT,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            upcast=True,
            downwind_index=downwind_index,
        )

        sp_sel = (ct > 1e-8) & (x <= 0)
        ct_sel = ct[sp_sel]
        r_sph_sel = r_sph[sp_sel]
        R_sel = D[sp_sel] / 2
        xi = r_sph_sel / R_sel

        def add_wake(sp_sel, wake_deltas, blockage):
            """adds to wake deltas"""
            if self.has_vector_wind_superp:
                wdeltas = {FV.WS: blockage}
                self.vec_superp.wdeltas_ws2uv(
                    algo, fdata, tdata, downwind_index, wdeltas, sp_sel
                )
                wake_deltas[FV.UV] = self.vec_superp.add_wake_vector(
                    algo,
                    mdata,
                    fdata,
                    tdata,
                    downwind_index,
                    sp_sel,
                    wake_deltas[FV.UV],
                    wdeltas.pop(FV.UV),
                )
            else:
                self.superp[FV.WS].add_wake(
                    algo,
                    mdata,
                    fdata,
                    tdata,
                    downwind_index,
                    sp_sel,
                    FV.WS,
                    wake_deltas[FV.WS],
                    blockage,
                )

        if np.any(sp_sel):
            blockage = self.induction.ct2a(ct_sel) * (1 + -xi / np.sqrt(1 + xi**2))
            add_wake(sp_sel, wake_deltas, -blockage)

        if not self.pre_rotor_only:
            sp_sel = (
                (ct > 1e-8) & (x > 0) & (r > D / 2)
            )  # mirror in rotor plane and inverse blockage, but not directly behind rotor
            ct_sel = ct[sp_sel]
            r_sph_sel = r_sph[sp_sel]
            R_sel = D[sp_sel] / 2
            xi = r_sph_sel / R_sel
            if np.any(sp_sel):
                blockage = self.induction.ct2a(ct_sel) * (1 + -xi / np.sqrt(1 + xi**2))
                add_wake(sp_sel, wake_deltas, blockage)

        return wake_deltas



"""
Static data collection.
"""

from .parse import parse_Pct_file_name as parse_Pct_file_name
from .parse import parse_Pct_two_files as parse_Pct_two_files

from .static_data import FARM as FARM
from .static_data import STATES as STATES
from .static_data import PCTCURVE as PCTCURVE
from .static_data import StaticData as StaticData



from foxes.core import TurbineType


class CalculatorType(TurbineType):
    """
    Direct data infusion by a user function.

    :group: models.turbine_types

    """

    def __init__(
        self,
        func,
        out_vars,
        *args,
        needs_rews2=False,
        needs_rews3=False,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        func: callable
            The function to calculate farm variables, should have the signature:
            f(algo, mdata, fdata, st_sel) -> dict, where the keys are
            output variable names and the values are numpy.ndarrays
            with shape (n_states, n_turbines).

            Beware that the turbine ordering in fdata is in downwind order,
            hence external data X of shape (n_states, n_turbines) in farm order
            needs to be reordered by X[ssel, order] with
            ssel = fdata[FV.ORDER_SSEL], order = fdata[FV.ORDER]
            before using it in combination with fdata variables.
        out_vars: list of str
            The output variables of the function
        args: tuple, optional
            Additional parameters for TurbineType class
        needs_rews2: bool
            Flag for runs that require the REWS2 variable
        needs_rews3: bool
            Flag for runs that require the REWS3 variable
        kwargs: dict, optional
            Additional parameters for TurbineType class

        """
        super().__init__(*args, **kwargs)
        self._func = func
        self._ovars = out_vars
        self._rews2 = needs_rews2
        self._rews3 = needs_rews3

    def needs_rews2(self):
        """
        Returns flag for requiring REWS2 variable

        Returns
        -------
        flag: bool
            True if REWS2 is required

        """
        return self._rews2

    def needs_rews3(self):
        """
        Returns flag for requiring REWS3 variable

        Returns
        -------
        flag: bool
            True if REWS3 is required

        """
        return self._rews3

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self._ovars

    def calculate(self, algo, mdata, fdata, st_sel):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        st_sel: numpy.ndarray of bool
            The state-turbine selection,
            shape: (n_states, n_turbines)

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        self.ensure_output_vars(algo, fdata)
        return self._func(algo, mdata, fdata, st_sel)



import numpy as np
from pathlib import Path
from sys import version_info

from foxes.utils.dict import Dict
from foxes.utils.load import import_module
import foxes.constants as FC


class Config(Dict):
    """
    Container for configurational data

    :group: foxes.config
    """

    def __init__(self):
        """Constructor"""
        super().__init__(
            {
                FC.DTYPE: np.float64,
                FC.ITYPE: np.int64,
                FC.WORK_DIR: Path("."),
                FC.INPUT_DIR: None,
                FC.OUTPUT_DIR: None,
                FC.NC_ENGINE: "netcdf4",
            },
            name="config",
        )
        self.__utmn = None
        self.__utml = None

        # special treat for Python 3.8:
        if version_info[0] == 3 and version_info[1] == 8:
            self["nc_engine"] = None

    def __setitem__(self, key, value):
        if key == FC.UTM_ZONE:
            raise KeyError(
                "Direct setting of UTM zone is not allowed. "
                "Use config.set_utm_zone(...) instead."
            )
        super().__setitem__(key, value)

    @property
    def dtype_double(self):
        """
        The default double data type

        Returns
        -------
        dtp: type
            The default double data type

        """
        return self.get_item(FC.DTYPE)

    @property
    def dtype_int(self):
        """
        The default int data type

        Returns
        -------
        dtp: type
            The default integer data type

        """
        return self.get_item(FC.ITYPE)

    @property
    def work_dir(self):
        """
        The foxes working directory

        Returns
        -------
        pth: pathlib.Path
            Path to the foxes working directory

        """
        pth = self.get_item(FC.WORK_DIR)
        if self[FC.WORK_DIR] is None:
            self[FC.WORK_DIR] = Path(".")
        elif not isinstance(pth, Path):
            self[FC.WORK_DIR] = Path(pth)
        return self[FC.WORK_DIR]

    @property
    def input_dir(self):
        """
        The input base directory

        Returns
        -------
        pth: pathlib.Path
            Path to the input base directory

        """
        if self[FC.INPUT_DIR] is None:
            return self.work_dir
        else:
            pth = self.get_item(FC.INPUT_DIR)
            if not isinstance(pth, Path):
                self[FC.INPUT_DIR] = Path(pth)
            return self[FC.INPUT_DIR]

    @property
    def output_dir(self):
        """
        The default output directory

        Returns
        -------
        pth: pathlib.Path
            Path to the default output directory

        """
        if self[FC.OUTPUT_DIR] is None:
            return self.work_dir
        else:
            pth = self.get_item(FC.OUTPUT_DIR)
            if not isinstance(pth, Path):
                self[FC.OUTPUT_DIR] = Path(pth)
            return self[FC.OUTPUT_DIR]

    @property
    def nc_engine(self):
        """
        The NetCDF engine

        Returns
        -------
        nce: str
            The NetCDF engine

        """
        nce = self[FC.NC_ENGINE]
        if nce == "netcdf4":
            import_module("netCDF4")
        elif nce is not None:
            import_module(nce)
        return nce

    @property
    def utm_zone_set(self):
        """
        Whether the UTM zone is set

        Returns
        -------
        uzs: bool
            True if both UTM zone number and letter are set

        """
        return self.__utmn is not None and self.__utml is not None

    @property
    def utm_zone(self):
        """
        The UTM zone (number, letter) tuple

        Returns
        -------
        zn: int
            The UTM zone number
        zl: str
            The UTM zone letter

        """
        assert self.utm_zone_set, "UTM zone has not been set"
        return self.__utmn, self.__utml

    def set_utm_zone(self, number, letter):
        """
        Set the UTM zone

        Parameters
        ----------
        number: int
            The UTM zone number
        letter: str
            The UTM zone letter
        verbosity: int
            The verbosity level, 0 = silent

        """
        assert not self.utm_zone_set, f"UTM zone already set to {self.utm_zone}"
        self.__utmn = number
        self.__utml = letter
        super().__setitem__(FC.UTM_ZONE, (number, letter))


config = Config()
"""Foxes configurational data object
:group: foxes.config
"""


def get_path(pth, base):
    """
    Gets path object, respecting the base directory

    Parameters
    ----------
    pth: str or pathlib.Path
        The path, optionally relative to base
    base: pathlib.Path
        The base directory

    Returns
    -------
    out: pathlib.Path
        The path, absolute or relative to base directory

    :group: foxes.config

    """
    if not isinstance(pth, Path):
        pth = Path(pth)
    if pth.is_absolute():
        return pth
    else:
        return base / pth


def get_input_path(pth):
    """
    Gets path object, respecting the configurations
    input directory

    Parameters
    ----------
    pth: str or pathlib.Path
        The path, optionally relative

    Returns
    -------
    out: pathlib.Path
        The path, absolute or relative to input directory
        from config

    :group: foxes.config

    """
    return get_path(pth, base=config.input_dir)


def get_output_path(pth):
    """
    Gets path object, respecting the configurations
    output directory

    Parameters
    ----------
    pth: str or pathlib.Path
        The path, optionally relative

    Returns
    -------
    out: pathlib.Path
        The path, absolute or relative to output directory
        from config

    :group: foxes.config

    """
    return get_path(pth, base=config.output_dir)



import time
import argparse
import matplotlib.pyplot as plt

import foxes
import foxes.variables as FV

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-l",
        "--layout",
        help="The wind farm layout file (path or static)",
        default="test_farm_67.csv",
    )
    parser.add_argument(
        "-s",
        "--states",
        help="The timeseries input file (path or static)",
        default="WRF-Timeseries-3000.nc",
    )
    parser.add_argument(
        "-t",
        "--turbine_file",
        help="The P-ct-curve csv file (path or static)",
        default="NREL-5MW-D126-H90.csv",
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="level10")
    parser.add_argument(
        "-p", "--pwakes", help="The partial wakes models", default=None, nargs="+"
    )
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["Jensen_linear_k007"],
        nargs="+",
    )
    parser.add_argument(
        "-m", "--tmodels", help="The turbine models", default=[], nargs="+"
    )
    parser.add_argument(
        "-sl",
        "--show_layout",
        help="Flag for showing layout figure",
        action="store_true",
    )
    parser.add_argument("-e", "--engine", help="The engine", default=None)
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-nf", "--nofig", help="Do not show figures", action="store_true"
    )
    args = parser.parse_args()

    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(args.turbine_file)
    mbook.turbine_types[ttype.name] = ttype

    states = foxes.input.states.MultiHeightNCTimeseries(
        data_source=args.states,
        time_coord="Time",
        h_coord="height",
        output_vars=[FV.WS, FV.WD, FV.TI, FV.RHO],
        var2col={FV.WS: "ws", FV.WD: "wd", FV.TI: "ti", FV.RHO: "rho"},
    )

    farm = foxes.WindFarm()
    foxes.input.farm_layout.add_from_file(
        farm, args.layout, turbine_models=args.tmodels + [ttype.name], H=170.0
    )

    if not args.nofig and args.show_layout:
        ax = foxes.output.FarmLayoutOutput(farm).get_figure()
        plt.show()
        plt.close(ax.get_figure())

    algo = foxes.algorithms.Downwind(
        farm,
        states,
        rotor_model=args.rotor,
        wake_models=args.wakes,
        wake_frame="rotor_wd",
        partial_wakes=args.pwakes,
        mbook=mbook,
    )

    with foxes.Engine.new(
        engine_type=args.engine,
        n_procs=args.n_cpus,
        chunk_size_states=args.chunksize_states,
    ):
        time0 = time.time()
        farm_results = algo.calc_farm()
        time1 = time.time()

    print("\nCalc time =", time1 - time0, "\n")

    print(farm_results, "\n")

    fr = farm_results.to_dataframe()
    print(fr[[FV.WD, FV.AMB_REWS, FV.REWS, FV.AMB_P, FV.P]])

    o = foxes.output.FarmResultsEval(farm_results)
    P0 = o.calc_mean_farm_power(ambient=True)
    P = o.calc_mean_farm_power()
    print(f"\nFarm power: {P / 1000:.1f} MW, Efficiency = {P / P0 * 100:.2f} %")



import numpy as np
from xarray import Dataset
from abc import abstractmethod
from tqdm import tqdm

from foxes.config import config, get_output_path
from foxes.core import Engine
from foxes.utils import write_nc as write_nc_file
import foxes.constants as FC


def _write_chunk_results(algo, results, write_nc, out_coords, mdata):
    """Helper function for optionally writing chunk results to netCDF file"""
    ret_data = True
    if write_nc is not None and write_nc["split"] == "chunks":
        ret_data = write_nc.get("ret_data", False)
        out_dir = get_output_path(write_nc.get("out_dir", "."))
        base_name = write_nc["base_name"]
        ret_data = write_nc.get("ret_data", False)
        out_dir.mkdir(parents=True, exist_ok=True)

        coords = {}
        if FC.STATE in out_coords and FC.STATE in mdata:
            coords[FC.STATE] = mdata[FC.STATE]

        dvars = {}
        for v, d in results.items():
            if (
                out_coords == (FC.STATE, FC.TURBINE)
                and d.shape[1] == 1
                and algo.n_turbines > 1
            ):
                dvars[v] = ((FC.STATE,), d[:, 0])
            else:
                dvars[v] = (out_coords, d)

        ds = Dataset(coords=coords, data_vars=dvars)

        i0 = mdata.chunki_states
        t0 = mdata.chunki_points
        vrb = max(algo.verbosity - 1, 0)
        if out_coords == (FC.STATE, FC.TURBINE):
            fpath = out_dir / f"{base_name}_{i0:04d}.nc"
        else:
            fpath = out_dir / f"{base_name}_{i0:04d}_{t0:04d}.nc"
        write_nc_file(ds, fpath, nc_engine=config.nc_engine, verbosity=vrb)

    return results if ret_data else None


def _run(
    algo,
    model,
    *data,
    iterative,
    chunk_store,
    chunk_key,
    out_coords,
    write_nc,
    **cpars,
):
    """Helper function for running in a single process"""
    algo.reset_chunk_store(chunk_store.copy())
    results = model.calculate(algo, *data, **cpars)
    chunk_store = algo.reset_chunk_store() if iterative else {}
    cstore = {chunk_key: chunk_store[chunk_key]} if chunk_key in chunk_store else {}
    results = _write_chunk_results(algo, results, write_nc, out_coords, data[0])
    return results, cstore


def _run_shared(
    algo,
    model,
    *data,
    chunk_key,
    out_coords,
    write_nc,
    **cpars,
):
    """Helper function for running in a single process"""
    results = model.calculate(algo, *data, **cpars)
    cstore = (
        {chunk_key: algo.chunk_store[chunk_key]}
        if chunk_key in algo.chunk_store
        else {}
    )
    results = _write_chunk_results(algo, results, write_nc, out_coords, data[0])
    return results, cstore


def _run_map(func, inputs, *args, **kwargs):
    """Helper function for running map func on proc"""
    return [func(x, *args, **kwargs) for x in inputs]


class PoolEngine(Engine):
    """
    Abstract engine for pool type parallelizations.

    Parameters
    ----------
    share_cstore: bool
        Share chunk store between chunks

    :group: engines

    """

    def __init__(self, *args, share_cstore=False, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        args: tuple, optional
            Arguments for the base class
        share_cstore: bool
            Share chunk store between chunks
        kwargs: dict, optional
            Additional arguments for the base class

        """
        super().__init__(*args, **kwargs)
        self.share_cstore = share_cstore

    @abstractmethod
    def _create_pool(self):
        """Creates the pool"""
        pass

    @abstractmethod
    def _shutdown_pool(self):
        """Shuts down the pool"""
        pass

    def __enter__(self):
        self._create_pool()
        return super().__enter__()

    def __exit__(self, *exit_args):
        self._shutdown_pool()
        super().__exit__(*exit_args)

    def map(
        self,
        func,
        inputs,
        *args,
        **kwargs,
    ):
        """
        Runs a function on a list of files

        Parameters
        ----------
        func: Callable
            Function to be called on each file,
            func(input, *args, **kwargs) -> data
        inputs: array-like
            The input data list
        args: tuple, optional
            Arguments for func
        kwargs: dict, optional
            Keyword arguments for func

        Returns
        -------
        results: list
            The list of results

        """
        if len(inputs) == 0:
            return []
        elif len(inputs) == 1:
            return [func(inputs[0], *args, **kwargs)]
        else:
            inptl = np.array_split(inputs, min(self.n_workers, len(inputs)))
            futures = []
            for subi in inptl:
                futures.append(self.submit(_run_map, func, subi, *args, **kwargs))
            results = []
            for f in futures:
                results += self.await_result(f)
            return results

    def run_calculation(
        self,
        algo,
        model,
        model_data,
        farm_data=None,
        point_data=None,
        out_vars=[],
        chunk_store={},
        sel=None,
        isel=None,
        iterative=False,
        write_nc=None,
        **calc_pars,
    ):
        """
        Runs the model calculation

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The algorithm object
        model: foxes.core.DataCalcModel
            The model that whose calculate function
            should be run
        model_data: xarray.Dataset
            The initial model data
        farm_data: xarray.Dataset, optional
            The initial farm data
        point_data: xarray.Dataset, optional
            The initial point data
        out_vars: list of str, optional
            Names of the output variables
        chunk_store: foxes.utils.Dict
            The chunk store
        sel: dict, optional
            Selection of coordinate subsets
        isel: dict, optional
            Selection of coordinate subsets index values
        iterative: bool
            Flag for use within the iterative algorithm
        write_nc: dict, optional
            Parameters for writing results to netCDF files, e.g.
            {'out_dir': 'results', 'base_name': 'calc_results',
            'ret_data': False, 'split': 1000}.

            The split parameter controls how the output is split:
            - 'chunks': one file per chunk (fastest method),
            - 'input': split according to sizes of multiple states input files,
            - int: split with this many states per file,
            - None: create a single output file.

            Use ret_data = False together with non-single file writing
            to avoid constructing the full Dataset in memory.
        calc_pars: dict, optional
            Additional parameters for the model.calculate()

        Returns
        -------
        results: xarray.Dataset
            The model results

        """

        # reset chunk store:
        if self.share_cstore:
            algo.reset_chunk_store(chunk_store)

        # subset selection:
        model_data, farm_data, point_data = self.select_subsets(
            model_data, farm_data, point_data, sel=sel, isel=isel
        )

        # basic checks:
        super().run_calculation(algo, model, model_data, farm_data, point_data)

        # prepare:
        n_states = model_data.sizes[FC.STATE]
        out_coords = model.output_coords()
        coords = {}
        if FC.STATE in out_coords and FC.STATE in model_data.coords:
            coords[FC.STATE] = model_data[FC.STATE].to_numpy()

        # DEBUG objec mem sizes:
        # from foxes.utils import print_mem
        # for m in [algo] + model.models:
        #    print_mem(m, pre_str="MULTIP CHECKING LARGE DATA", min_csize=9999)

        # calculate chunk sizes:
        n_targets = point_data.sizes[FC.TARGET] if point_data is not None else 0
        chunk_sizes_states, chunk_sizes_targets = self.calc_chunk_sizes(
            n_states, n_targets
        )
        n_chunks_states = len(chunk_sizes_states)
        n_chunks_targets = len(chunk_sizes_targets)
        self.print(
            f"{type(self).__name__}: Selecting n_chunks_states = {n_chunks_states}, n_chunks_targets = {n_chunks_targets}",
            level=2,
        )

        # prepare and submit chunks:
        n_chunks_all = n_chunks_states * n_chunks_targets
        self.print(
            f"{type(self).__name__}: Submitting {n_chunks_all} chunks to {self.n_workers} workers",
            level=2,
        )
        pbar = None
        if self.verbosity > 1 and self.has_progress_bar:
            pbar = tqdm(total=n_chunks_all)
        futures = {}
        i0_states = 0
        pdone = -1
        counter = 0
        for chunki_states in range(n_chunks_states):
            i1_states = i0_states + chunk_sizes_states[chunki_states]
            i0_targets = 0
            for chunki_points in range(n_chunks_targets):
                key = (chunki_states, chunki_points)
                i1_targets = i0_targets + chunk_sizes_targets[chunki_points]

                # get this chunk's data:
                data = self.get_chunk_input_data(
                    algo=algo,
                    model_data=model_data,
                    farm_data=farm_data,
                    point_data=point_data,
                    states_i0_i1=(i0_states, i1_states),
                    targets_i0_i1=(i0_targets, i1_targets),
                    out_vars=out_vars,
                    chunki_states=chunki_states,
                    chunki_points=chunki_points,
                    n_chunks_states=n_chunks_states,
                    n_chunks_points=n_chunks_targets,
                )

                # submit model calculation:
                if self.share_cstore:
                    futures[(chunki_states, chunki_points)] = self.submit(
                        _run_shared,
                        algo,
                        model,
                        *data,
                        chunk_key=key,
                        out_coords=out_coords,
                        write_nc=write_nc,
                        **calc_pars,
                    )
                else:
                    futures[(chunki_states, chunki_points)] = self.submit(
                        _run,
                        algo,
                        model,
                        *data,
                        iterative=iterative,
                        chunk_store=chunk_store,
                        chunk_key=key,
                        out_coords=out_coords,
                        write_nc=write_nc,
                        **calc_pars,
                    )
                del data

                i0_targets = i1_targets

                counter += 1
                if pbar is not None:
                    pbar.update()
                elif self.verbosity > 1 and self.prints_progress and n_chunks_all > 1:
                    pr = int(100 * counter / n_chunks_all)
                    if pr > pdone:
                        pdone = pr
                        print(
                            f"{self.name}: Submitted {counter} of {n_chunks_all} chunks, {pdone}%"
                        )
            i0_states = i1_states

        if farm_data is None:
            farm_data = Dataset()
        goal_data = farm_data if point_data is None else point_data
        del calc_pars, farm_data, point_data

        if pbar is not None:
            pbar.close()
        elif self.verbosity > 1 and self.prints_progress:
            print(f"{type(self).__name__}: Submitted all {counter} chunks\n")

        return self.combine_results(
            algo=algo,
            futures=futures,
            model_data=model_data,
            out_vars=out_vars,
            out_coords=out_coords,
            n_chunks_states=n_chunks_states,
            n_chunks_targets=n_chunks_targets,
            goal_data=goal_data,
            iterative=iterative,
            write_nc=write_nc,
        )



from abc import abstractmethod
import numpy as np

from foxes.utils import new_instance, wd2uv, uv2wd
from foxes.config import config
import foxes.variables as FV
import foxes.constants as FC

from .model import Model
from .data import TData


class PartialWakesModel(Model):
    """
    Abstract base class for partial wakes models.

    Partial wakes models compute wake effects
    for rotor effective quantities.

    Attributes
    ----------
    wake_models: list of foxes.core.WakeModel
        The wake model selection
    wake_frame: foxes.core.WakeFrame, optional
        The wake frame

    :group: core

    """

    def check_wmodel(self, wmodel, error=True):
        """
        Checks the wake model type

        Parameters
        ----------
        wmodel: foxes.core.WakeModel
            The wake model to be tested
        error: bool
            Flag for raising TypeError

        Returns
        -------
        chk: bool
            True if wake model is compatible

        """
        return True

    @abstractmethod
    def get_wake_points(self, algo, mdata, fdata):
        """
        Get the wake calculation points, and their
        weights.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data

        Returns
        -------
        rpoints: numpy.ndarray
            The wake calculation points, shape:
            (n_states, n_turbines, n_tpoints, 3)
        rweights: numpy.ndarray
            The target point weights, shape: (n_tpoints,)

        """
        pass

    def get_initial_tdata(
        self,
        algo,
        mdata,
        fdata,
        amb_rotor_res,
        rotor_weights,
        wmodels,
    ):
        """
        Creates the initial target data object

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        amb_rotor_res: dict
            The ambient results at rotor points,
            key: variable name, value: numpy.ndarray
            of shape: (n_states, n_turbines, n_rotor_points)
        rotor_weights: numpy.ndarray
            The rotor point weights, shape: (n_rotor_points,)
        wmodels: list of foxes.core.WakeModel
            The wake models for this partial wake model

        Returns
        -------
        tdata: foxes.core.TData
            The target point data for the wake points

        """
        tpoints, tweights = self.get_wake_points(algo, mdata, fdata)
        tdata = TData.from_tpoints(tpoints, tweights)

        # map wind data:
        if FV.WD in amb_rotor_res or FV.WS in amb_rotor_res:
            assert FV.WD in amb_rotor_res and FV.WS in amb_rotor_res, (
                "Require both wind direction and speed in ambient rotor results."
            )
            uv = wd2uv(amb_rotor_res[FV.WD], amb_rotor_res[FV.WS])
            uv = np.stack(
                [
                    self.map_rotor_results(
                        algo, mdata, fdata, tdata, FV.U, uv[..., 0], rotor_weights
                    ),
                    self.map_rotor_results(
                        algo, mdata, fdata, tdata, FV.V, uv[..., 1], rotor_weights
                    ),
                ],
                axis=-1,
            )
            tdata.add(FV.AMB_WD, uv2wd(uv), dims=(FC.STATE, FC.TARGET, FC.TPOINT))
            tdata.add(
                FV.AMB_WS,
                np.linalg.norm(uv, axis=-1),
                dims=(FC.STATE, FC.TARGET, FC.TPOINT),
            )
            for wmodel in wmodels:
                if wmodel.has_uv:
                    tdata.add(
                        FV.AMB_UV, uv, dims=(FC.STATE, FC.TARGET, FC.TPOINT, FC.XY)
                    )
                    break

        # map rotor point results onto target points:
        for v, d in amb_rotor_res.items():
            if v not in [FV.WS, FV.WD, FV.U, FV.V, FV.UV]:
                w = FV.var2amb.get(v, v)
                tdata.add(
                    w,
                    self.map_rotor_results(
                        algo, mdata, fdata, tdata, v, d, rotor_weights
                    ),
                    dims=(FC.STATE, FC.TARGET, FC.TPOINT),
                )

        return tdata

    def map_rotor_results(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        variable,
        rotor_res,
        rotor_weights,
    ):
        """
        Map ambient rotor point results onto target points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        variable: str
            The variable name to map
        rotor_res: numpy.ndarray
            The results at rotor points, shape:
            (n_states, n_turbines, n_rotor_points)
        rotor_weights: numpy.ndarray
            The rotor point weights, shape: (n_rotor_points,)

        Returns
        -------
        res: numpy.ndarray
            The mapped results at target points, shape:
            (n_states, n_targets, n_tpoints)

        """
        if len(rotor_res.shape) > 2 and rotor_res.shape[:2] == (
            tdata.n_states,
            tdata.n_targets,
        ):
            q = np.zeros(
                (tdata.n_states, tdata.n_targets, tdata.n_tpoints),
                dtype=config.dtype_double,
            )
            if rotor_res.shape[2] == 1:
                q[:] = rotor_res
            else:
                q[:] = np.einsum("str,r->st", rotor_res, rotor_weights)[:, :, None]
            return q
        else:
            raise ValueError(
                f"Partial wakes '{self.name}': Incompatible shape '{rotor_res.shape}' for variable '{variable}' in rotor results."
            )

    def new_wake_deltas(self, algo, mdata, fdata, tdata, wmodel):
        """
        Creates new initial wake deltas, filled
        with zeros.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        wmodel: foxes.core.WakeModel
            The wake model

        Returns
        -------
        wake_deltas: dict
            Key: variable name, value: The zero filled
            wake deltas, shape: (n_states, n_turbines, n_tpoints, ...)

        """
        return wmodel.new_wake_deltas(algo, mdata, fdata, tdata)

    def contribute(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        wake_deltas,
        wmodel,
    ):
        """
        Modifies wake deltas at target points by
        contributions from the specified wake source turbines.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints, ...)
        wmodel: foxes.core.WakeModel
            The wake model

        """
        wcoos = algo.wake_frame.get_wake_coos(algo, mdata, fdata, tdata, downwind_index)
        wmodel.contribute(algo, mdata, fdata, tdata, downwind_index, wcoos, wake_deltas)

    @abstractmethod
    def finalize_wakes(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        rpoint_weights,
        wake_deltas,
        wmodel,
        downwind_index,
    ):
        """
        Updates the wake_deltas at the selected target
        downwind index.

        Modifies wake_deltas on the fly.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.Data
            The target point data
        rpoint_weights: numpy.ndarray
            The rotor point weights, shape: (n_rotor_points,)
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: np.ndarray of shape
            (n_states, n_turbines, n_tpoints)
        wmodel: foxes.core.WakeModel
            The wake model
        downwind_index: int
            The index in the downwind order

        Returns
        -------
        final_wake_deltas: dict
            The final wake deltas at the selected downwind
            turbines. Key: variable name, value: np.ndarray
            of shape (n_states, n_rotor_points)

        """
        pass

    @classmethod
    def new(cls, pwakes_type, *args, **kwargs):
        """
        Run-time partial wakes model factory.

        Parameters
        ----------
        pwakes_type: str
            The selected derived class name
        args: tuple, optional
            Additional parameters for the constructor
        kwargs: dict, optional
            Additional parameters for the constructor

        """
        return new_instance(cls, pwakes_type, *args, **kwargs)



"""
Point models.
"""

from .wake_deltas import WakeDeltas as WakeDeltas
from .set_uniform_data import SetUniformData as SetUniformData
from .tke2ti import TKE2TI as TKE2TI
from .ustar2ti import Ustar2TI as Ustar2TI



import matplotlib.pyplot as plt
import pandas as pd
from inspect import signature

import foxes.input.farm_layout as farm_layout
from foxes.core import States, Engine, WindFarm, Algorithm
from foxes.models import ModelBook
from foxes.output import Output
from foxes.utils import Dict, new_cls
from foxes.config import config
import foxes.constants as FC


def read_dict(
    idict,
    farm=None,
    states=None,
    mbook=None,
    algo=None,
    engine_pars=None,
    iterative=None,
    verbosity=None,
    work_dir=None,
    input_dir=None,
    output_dir=None,
    **algo_pars,
):
    """
    Read dictionary input into foxes objects

    Parameters
    ----------
    idict: foxes.utils.Dict
        The input parameter dictionary
    farm: foxes.core.WindFarm, optional
        The wind farm, overrules settings from idict
    states: foxes.core.States, optional
        The ambient states, overrules settings from idict
    mbook: foxes.models.ModelBook, optional
        The model book, overrules settings from idict
    algo: foxes.core.Algorithm, optional
        The algorithm, overrules settings from idict
    engine_pars: dict, optional
        Parameters for engine creation, overrules
        settings from idict
    iterative: bool, optional
        Force iterative calculations, overrules
        settings from idict
    verbosity: int, optional
        Force a verbosity level, 0 = silent, overrules
        settings from idict
    work_dir: str or pathlib.Path, optional
        Path to the working directory
    input_dir: str or pathlib.Path, optional
        The default input directory
    output_dir: str or pathlib.Path, optional
        The default output directory
    algo_pars: dict, optional
        Additional parameters for the algorithm, overrules
        settings from idict

    Returns
    -------
    algo: foxes.core.Algorithm
        The algorithm
    engine: foxes.core.Engine
        The engine, or None if not set

    :group: input.yaml

    """

    def _print(*args, level=1, **kwargs):
        if verbosity is None or verbosity >= level:
            print(*args, **kwargs)

    # set working directory:
    ld = 0
    for c, d in zip(
        [FC.WORK_DIR, FC.INPUT_DIR, FC.OUTPUT_DIR], [work_dir, input_dir, output_dir]
    ):
        if d is not None:
            config[c] = d
            ld = max(ld, len(str(d)))
    _print("\n--------------------- Reading foxes parameter dict ---------------------")
    _print("Working directory  :", config.work_dir)
    _print("Input directory    :", config.input_dir)
    _print("Output directory   :", config.output_dir)

    # create states:
    if states is None:
        if algo is None:
            _print("Creating states")
            states = States.new(**idict["states"])
        else:
            states = algo.states
    else:
        assert algo is None, (
            "Cannot handle both the algo and the states argument, please drop one"
        )

    # create model book:
    if mbook is None:
        if algo is None:
            mbook = ModelBook()
            if "model_book" in idict:
                _print("Creating model book")
                mdict = idict.get_item("model_book")
                for s, mlst in mdict.items():
                    t = mbook.sources.get_item(s)
                    c = mbook.base_classes.get_item(s)
                    ms = [
                        Dict(m, _name=f"{mdict.name}.s.{i}") for i, m in enumerate(mlst)
                    ]
                    for m in ms:
                        mname = m.pop_item("name")
                        _print(f"  Adding {s}.{mname}")
                        t[mname] = c.new(**m)
        else:
            mbook = algo.mbook
    else:
        assert algo is None, (
            "Cannot handle both the algo and the mbook argument, please drop one"
        )

    # create farm:
    if farm is None:
        if algo is None:
            _print("Creating wind farm")
            fdict = idict.get_item("wind_farm")
            lyts = fdict.pop_item("layouts")
            farm = WindFarm(**fdict)
            for lyt in lyts:
                add_fun = getattr(farm_layout, lyt.pop_item("function"))
                if verbosity is not None:
                    lyt["verbosity"] = verbosity - 1
                add_fun(farm, **lyt)
        else:
            farm = algo.farm
    else:
        assert algo is None, (
            "Cannot handle both the algo and the farm argument, please drop one"
        )

    # create engine:
    engine = None
    if engine_pars is not None:
        engine = Engine.new(**engine_pars)
        _print(f"Initializing engine: {engine}")
        engine.initialize()
    elif "engine" in idict:
        if verbosity is not None:
            idict["verbosity"] = verbosity - 1
        engine = Engine.new(**idict["engine"])
        engine.initialize()
        _print(f"Initializing engine: {engine}")

    # create algorithm:
    if algo is None:
        adict = idict.get_item("algorithm")
        if iterative is not None and iterative:
            adict["algo_type"] = "Iterative"
        _print("Creating algorithm :", adict["algo_type"])
        adict.update(dict(farm=farm, states=states, mbook=mbook))
        if verbosity is not None:
            adict["verbosity"] = verbosity - 1
        if algo_pars is not None:
            adict.update({v: d for v, d in algo_pars.items() if d is not None})
        algo = Algorithm.new(**adict)

    _print("------------------------------------------------------------------------\n")

    return algo, engine


def get_output_obj(
    ocls,
    odict,
    algo,
    farm_results=None,
    point_results=None,
    base_class=Output,
    extra_sig={},
):
    """
    Create the output object

    Parameters
    ----------
    ocls: str
        Name of the output class
    odict: dict
        The output dict
    algo: foxes.core.Algorithm
        The algorithm
    farm_results: xarray.Dataset, optional
        The farm results
    point_results: xarray.Dataset, optional
        The point results
    base_class: object
        The output's base class
    extra_sig: dict
        Extra function signature check, sets
        arguments (key) with data (value)

    Returns
    -------
    obj: object or None
        The output object

    :group: input.yaml

    """
    cls = new_cls(base_class, ocls)
    prs = list(signature(cls.__init__).parameters.keys())
    if "algo" in prs:
        assert algo is not None, f"Output of type '{ocls}' requires algo"
        odict["algo"] = algo
    if "farm" in prs:
        odict["farm"] = algo.farm
    if "farm_results" in prs:
        if farm_results is None:
            print(f"No farm results; skipping output {ocls}")
            return None
        odict["farm_results"] = farm_results
    if "point_results" in prs:
        odict["point_results"] = point_results
    for k, v in extra_sig.items():
        if k in prs:
            odict[k] = v

    return cls(**odict)


def _get_object(rlabels, d):
    """Helper function for object extraction"""
    d = d.replace("]", "")
    i0 = d.find("[")
    if i0 > 0:
        inds = tuple([int(x) for x in d[i0 + 1 :].split(",")])
        return rlabels[d[:i0]][inds]
    else:
        return rlabels[d]


def run_obj_function(
    obj,
    fdict,
    algo,
    rlabels,
    nofig=False,
    verbosity=None,
):
    """
    Runs a function of an object

    Parameters
    ----------
    obj: object
        The object
    fdict: dict
        The function call dict
    algo: foxes.core.Algorithm
        The algorithm
    rlabels: dict
        Storage for result variables
    nofig: bool
        Do not show figures, overrules settings from fdict
    verbosity: int, optional
        The verbosity level, 0 = silent

    Returns
    -------
    results: object
        The returns of the function

    :group: input.yaml

    """

    def _print(*args, level=1, **kwargs):
        if verbosity is None or verbosity >= level:
            print(*args, **kwargs)

    fname = fdict.pop_item("function")
    _print(f"Running function {type(obj).__name__}.{fname}")
    plt_show = fdict.pop_item("plt_show", False)
    plt_close = fdict.pop_item("plt_close", False)
    rlbs = fdict.pop_item("result_labels", None)

    # grab function:
    ocls = type(obj).__name__
    assert hasattr(obj, fname), f"Output of type '{ocls}': Function '{fname}' not found"
    f = getattr(obj, fname)

    # add required input data objects:
    prs = list(signature(f).parameters.keys())
    if "algo" in prs:
        fdict["algo"] = algo
    if "farm" in prs:
        fdict["farm"] = algo.farm

    # replace result labels by objects:
    for k, d in fdict.items():
        if isinstance(d, str) and d[0] == "$":
            fdict[k] = _get_object(rlabels, d)

    # run function:
    args = fdict.pop_item("args", tuple())
    results = f(*args, **fdict)

    # pyplot shortcuts:
    if not nofig and plt_show:
        plt.show()
    if not nofig and plt_close:
        results = None
        plt.close()

    # store results under result labels:
    if rlbs is not None:

        def _set_label(rlabels, k, r):
            if k not in ["", "none", "None", "_", "__"]:
                assert k[0] == "$", (
                    f"Output {i} of type '{ocls}', function '{fname}': result labels must start with '$', got '{k}'"
                )
                assert "[" not in k and "]" not in k and "," not in k, (
                    f"Output {i} of type '{ocls}', function '{fname}': result labels cannot contain '[' or ']' or comma, got '{k}'"
                )
                _print(f"    result label {k}: {type(r).__name__}")
                rlabels[k] = r

        if isinstance(rlbs, (list, tuple)):
            for i, k in enumerate(rlbs):
                _set_label(rlabels, k, results[i])
        else:
            _set_label(rlabels, rlbs, results)

    return results


def run_outputs(
    idict,
    algo=None,
    farm_results=None,
    point_results=None,
    extra_sig={},
    ret_rlabels=False,
    nofig=False,
    verbosity=None,
):
    """
    Run outputs from dict.

    Parameters
    ----------
    idict: foxes.utils.Dict
        The input parameter dictionary
    algo: foxes.core.Algorithm, optional
        The algorithm
    farm_results: xarray.Dataset, optional
        The farm results
    point_results: xarray.Dataset, optional
        The point results
    extra_sig: dict
        Extra function signature check, sets
        arguments (key) with data (value)
    ret_rlabels: bool
        Flag for returning results variables
    nofig: bool
        Do not show figures, overrules settings from idict
    verbosity: int, optional
        The verbosity level, 0 = silent

    Returns
    -------
    outputs: list of tuple
        For each output enty, a tuple (dict, results),
        where results is a list that represents one
        entry per function call
    rlabels: dict, optional
        The results variables

    :group: input.yaml

    """

    def _print(*args, level=1, **kwargs):
        if verbosity is None or verbosity >= level:
            print(*args, **kwargs)

    out = []
    rlabels = Dict(_name="result_labels")
    if "outputs" in idict:
        odicts = idict["outputs"]

        for i, d in enumerate(odicts):
            if "output_type" in d:
                d["nofig"] = nofig
                ocls = d.pop_item("output_type")
                _print(f"\nRunning output {i}: {ocls}")
                d0 = dict(output_type=ocls)
                d0.update(d)

                flist = d.pop_item("functions")

                o = get_output_obj(
                    ocls, d, algo, farm_results, point_results, extra_sig=extra_sig
                )
                if o is None:
                    out.append((d0, None))
                    continue

            elif "object" in d:
                ocls = d.pop_item("object")
                _print(f"Running output {i}: Object {ocls}")
                o = _get_object(rlabels, ocls)
                d0 = dict(object=ocls)
                d0.update(d)
                flist = d.pop_item("functions")

            else:
                raise KeyError(
                    f"Output {i}: Please specify either 'output_type' or 'object'"
                )

            fres = []
            for fdict in flist:
                results = run_obj_function(o, fdict, algo, rlabels, nofig, verbosity)
                fres.append(results)
            out.append((d0, fres))

        if len(odicts):
            _print()

    return out if not ret_rlabels else out, rlabels


def run_dict(idict, *args, nofig=False, verbosity=None, **kwargs):
    """
    Runs foxes from dictionary input

    Parameters
    ----------
    idict: foxes.utils.Dict
        The input parameter dictionary
    args: tuple, optional
        Additional parameters for read_dict
    nofig: bool
        Do not show figures, overrules settings from idict
    verbosity: int, optional
        Force a verbosity level, 0 = silent, overrules
        settings from idict
    kwargs: dict, optional
        Additional parameters for read_dict

    Returns
    -------
    farm_results: xarray.Dataset, optional
        The farm results
    point_results: xarray.Dataset, optional
        The point results
    outputs: list of tuple
        For each output enty, a tuple (dict, results),
        where results is a list that represents one
        entry per function call

    :group: input.yaml

    """

    def _print(*args, level=1, **kwargs):
        if verbosity is None or verbosity >= level:
            print(*args, **kwargs)

    # read components:
    algo, engine = read_dict(idict, *args, verbosity=verbosity, **kwargs)

    # run farm calculation:
    rdict = idict.get_item("calc_farm", Dict(_name=idict.name + ".calc_farm"))
    if rdict.pop_item("run", True):
        _print("Running calc_farm")
        farm_results = algo.calc_farm(**rdict)
    else:
        farm_results = None
    out = (farm_results,)

    # run points calculation:
    point_results = None
    if "calc_points" in idict:
        rdict = idict.get_item("calc_points")
        if rdict.pop_item("run"):
            _print("Running calc_points")
            points = rdict.pop_item("points")
            if isinstance(points, str):
                _print("Reading file", points)
                points = pd.read_csv(points).to_numpy()
            point_results = algo.calc_points(farm_results, points=points, **rdict)
        else:
            point_results = None
        out += (point_results,)

    # run outputs:
    out += (
        run_outputs(
            idict, algo, farm_results, point_results, nofig=nofig, verbosity=verbosity
        ),
    )

    # shutdown engine, if created above:
    if engine is not None:
        _print(f"Finalizing engine: {engine}")
        engine.finalize()

    return out



import numpy as np
import xarray as xr
from copy import deepcopy
from tqdm import tqdm

from foxes.core import Engine, MData, FData, TData
from foxes.utils import import_module
import foxes.constants as FC

from .pool import _run_shared

dask = None
distributed = None


def delayed(func):
    """A dummy decorator"""
    return func


def load_dask():
    """On-demand loading of the dask package"""
    global dask, ProgressBar, delayed
    if dask is None:
        dask = import_module("dask")
        ProgressBar = import_module(
            "dask.diagnostics",
            pip_hint="pip install dask",
            conda_hint="conda install dask -c conda-forge",
        ).ProgressBar
        delayed = dask.delayed


def load_distributed():
    """On-demand loading of the distributed package"""
    global distributed
    if distributed is None:
        distributed = import_module("distributed")


@delayed
def _run_map(func, inputs, *args, **kwargs):
    """Helper function for running map func on proc"""
    return [func(x, *args, **kwargs) for x in inputs]


class DaskBaseEngine(Engine):
    """
    Abstract base class for foxes calculations with dask.

    Parameters
    ----------
    dask_config: dict
        The dask configuration parameters

    :group: engines

    """

    def __init__(
        self,
        *args,
        dask_config={},
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        args: tuple, optional
            Additional parameters for the base class
        dask_config: dict, optional
            The dask configuration parameters
        progress_bar: bool
            Flag for showing progress bar
        kwargs: dict, optional
            Additional parameters for the base class

        """
        super().__init__(*args, **kwargs)

        load_dask()

        self.dask_config = dask_config

    def __enter__(self):
        if self.has_progress_bar:
            self._pbar = ProgressBar(minimum=2)
            self._pbar.__enter__()
        return super().__enter__()

    def __exit__(self, *args):
        if self.has_progress_bar:
            self._pbar.__exit__(*args)
        super().__exit__(*args)

    def initialize(self):
        """
        Initializes the engine.
        """
        dask.config.set(**self.dask_config)
        super().initialize()

    def submit(self, f, *args, **kwargs):
        """
        Submits a job to worker, obtaining a future

        Parameters
        ----------
        f: Callable
            The function f(*args, **kwargs) to be
            submitted
        args: tuple, optional
            Arguments for the function
        kwargs: dict, optional
            Arguments for the function

        Returns
        -------
        future: object
            The future object

        """
        return delayed(f)(*args, **kwargs)

    def await_result(self, future):
        """
        Waits for result from a future

        Parameters
        ----------
        future: object
            The future

        Returns
        -------
        result: object
            The calculation result

        """
        return future.compute()

    def map(
        self,
        func,
        inputs,
        *args,
        **kwargs,
    ):
        """
        Runs a function on a list of files

        Parameters
        ----------
        func: Callable
            Function to be called on each file,
            func(input, *args, **kwargs) -> data
        inputs: array-like
            The input data list
        args: tuple, optional
            Arguments for func
        kwargs: dict, optional
            Keyword arguments for func

        Returns
        -------
        results: list
            The list of results

        """
        if len(inputs) == 0:
            return []
        elif len(inputs) == 1:
            return [func(inputs[0], *args, **kwargs)]
        else:
            inptl = np.array_split(inputs, min(self.n_workers, len(inputs)))
            futures = []
            for subi in inptl:
                futures.append(_run_map(func, subi, *args, **kwargs))
            results = dask.compute(futures)[0]
            out = []
            for r in results:
                out += r
            return out

    def chunk_data(self, data):
        """
        Applies the selected chunking

        Parameters
        ----------
        data: xarray.Dataset
            The data to be chunked

        Returns
        -------
        data: xarray.Dataset
            The chunked data

        """
        cks = {}
        cks[FC.STATE] = min(data.sizes[FC.STATE], self.chunk_size_states)
        if FC.TARGET in data.sizes:
            cks[FC.TARGET] = min(data.sizes[FC.TARGET], self.chunk_size_points)

        if len(set(cks.keys()).intersection(data.coords.keys())):
            return data.chunk({v: d for v, d in cks.items() if v in data.coords})
        else:
            return data

    def finalize(self, *exit_args, **exit_kwargs):
        """
        Finalizes the engine.

        Parameters
        ----------
        exit_args: tuple, optional
            Arguments from the exit function
        exit_kwargs: dict, optional
            Arguments from the exit function

        """
        dask.config.refresh()
        super().finalize(*exit_args, **exit_kwargs)


class DaskEngine(DaskBaseEngine):
    """
    The dask engine for delayed foxes calculations.

    :group: engines

    """

    def run_calculation(
        self,
        algo,
        model,
        model_data=None,
        farm_data=None,
        point_data=None,
        out_vars=[],
        chunk_store={},
        sel=None,
        isel=None,
        iterative=False,
        write_nc=None,
        **calc_pars,
    ):
        """
        Runs the model calculation

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The algorithm object
        model: foxes.core.DataCalcModel
            The model that whose calculate function
            should be run
        model_data: xarray.Dataset
            The initial model data
        farm_data: xarray.Dataset
            The initial farm data
        point_data: xarray.Dataset
            The initial point data
        out_vars: list of str, optional
            Names of the output variables
        chunk_store: foxes.utils.Dict
            The chunk store
        sel: dict, optional
            Selection of coordinate subsets
        isel: dict, optional
            Selection of coordinate subsets index values
        iterative: bool
            Flag for use within the iterative algorithm
        write_nc: dict, optional
            Parameters for writing results to netCDF files, e.g.
            {'out_dir': 'results', 'base_name': 'calc_results',
            'ret_data': False, 'split': 1000}.

            The split parameter controls how the output is split:
            - 'chunks': one file per chunk (fastest method),
            - 'input': split according to sizes of multiple states input files,
            - int: split with this many states per file,
            - None: create a single output file.

            Use ret_data = False together with non-single file writing
            to avoid constructing the full Dataset in memory.
        calc_pars: dict, optional
            Additional parameters for the model.calculate()

        Returns
        -------
        results: xarray.Dataset
            The model results

        """
        # reset chunk store:
        algo.reset_chunk_store(chunk_store)

        # subset selection:
        model_data, farm_data, point_data = self.select_subsets(
            model_data, farm_data, point_data, sel=sel, isel=isel
        )

        # basic checks:
        super().run_calculation(algo, model, model_data, farm_data, point_data)

        # prepare:
        n_states = model_data.sizes[FC.STATE]
        out_coords = model.output_coords()
        coords = {}
        if FC.STATE in out_coords and FC.STATE in model_data.coords:
            coords[FC.STATE] = model_data[FC.STATE].to_numpy()
        if farm_data is None:
            farm_data = xr.Dataset()
        goal_data = farm_data if point_data is None else point_data
        algo.reset_chunk_store(chunk_store)

        # calculate chunk sizes:
        n_targets = point_data.sizes[FC.TARGET] if point_data is not None else 0
        chunk_sizes_states, chunk_sizes_targets = self.calc_chunk_sizes(
            n_states, n_targets
        )
        n_chunks_states = len(chunk_sizes_states)
        n_chunks_targets = len(chunk_sizes_targets)
        self.print(
            f"Selecting n_chunks_states = {n_chunks_states}, n_chunks_targets = {n_chunks_targets}",
            level=2,
        )

        # submit chunks:
        n_chunks_all = n_chunks_states * n_chunks_targets
        self.print(
            f"Submitting {n_chunks_all} chunks to {self.n_workers} workers", level=2
        )
        pbar = (
            tqdm(total=n_chunks_all)
            if self.verbosity > 1 and self.has_progress_bar
            else None
        )
        futures = {}
        i0_states = 0
        counter = 0
        pdone = -1
        for chunki_states in range(n_chunks_states):
            i1_states = i0_states + chunk_sizes_states[chunki_states]
            i0_targets = 0
            for chunki_points in range(n_chunks_targets):
                key = (chunki_states, chunki_points)
                i1_targets = i0_targets + chunk_sizes_targets[chunki_points]

                # get this chunk's data:
                data = self.get_chunk_input_data(
                    algo=algo,
                    model_data=model_data,
                    farm_data=farm_data,
                    point_data=point_data,
                    states_i0_i1=(i0_states, i1_states),
                    targets_i0_i1=(i0_targets, i1_targets),
                    out_vars=out_vars,
                    chunki_states=chunki_states,
                    chunki_points=chunki_points,
                    n_chunks_states=n_chunks_states,
                    n_chunks_points=n_chunks_targets,
                )

                # submit model calculation:
                futures[(chunki_states, chunki_points)] = delayed(_run_shared)(
                    algo,
                    model,
                    *data,
                    chunk_key=key,
                    out_coords=out_coords,
                    write_nc=write_nc,
                    **calc_pars,
                )
                del data

                i0_targets = i1_targets

                counter += 1
                if pbar is not None:
                    pbar.update()
                elif self.verbosity > 1 and self.prints_progress:
                    pr = int(100 * counter / n_chunks_states)
                    if pr > pdone:
                        pdone = pr
                        print(
                            f"{self.name}: Submitted {counter} of {n_chunks_states} chunks, {pdone}%"
                        )
            i0_states = i1_states

        del farm_data, point_data, calc_pars
        if pbar is not None:
            pbar.close()

        # wait for results:
        if n_chunks_all > 1 or self.verbosity > 1:
            self.print(
                f"Computing {n_chunks_all} chunks using {self.n_workers} workers"
            )
        results = dask.compute(futures)[0]
        del futures

        return self.combine_results(
            algo=algo,
            futures=None,
            results=results,
            model_data=model_data,
            out_vars=out_vars,
            out_coords=out_coords,
            n_chunks_states=n_chunks_states,
            n_chunks_targets=n_chunks_targets,
            goal_data=goal_data,
            iterative=iterative,
            write_nc=write_nc,
        )


def _run_on_cluster(
    algo,
    model,
    *data,
    names,
    dims,
    mdata_size,
    fdata_size,
    iterative,
    chunk_store,
    chunki_states,
    chunki_points,
    n_chunks_states,
    n_chunks_points,
    i0_states,
    cpars,
):
    """Helper function for running on a cluster"""

    algo.reset_chunk_store(chunk_store)

    mdata = MData(
        data={names[i]: data[i] for i in range(mdata_size)},
        dims={names[i]: dims[i] for i in range(mdata_size)},
        chunki_states=chunki_states,
        chunki_points=chunki_points,
        n_chunks_states=n_chunks_states,
        n_chunks_points=n_chunks_points,
        states_i0=i0_states,
    )

    fdata_end = mdata_size + fdata_size
    fdata = FData(
        data={names[i]: data[i].copy() for i in range(mdata_size, fdata_end)},
        dims={names[i]: dims[i] for i in range(mdata_size, fdata_end)},
        chunki_states=chunki_states,
        chunki_points=chunki_points,
        n_chunks_states=n_chunks_states,
        n_chunks_points=n_chunks_points,
        states_i0=i0_states,
    )

    tdata = None
    if len(data) > fdata_end:
        tdata = TData(
            data={names[i]: data[i].copy() for i in range(fdata_end, len(data))},
            dims={names[i]: dims[i] for i in range(fdata_end, len(data))},
            chunki_states=chunki_states,
            chunki_points=chunki_points,
            n_chunks_states=n_chunks_states,
            n_chunks_points=n_chunks_points,
            states_i0=i0_states,
        )

    data = [d for d in [mdata, fdata, tdata] if d is not None]

    results = model.calculate(algo, *data, **cpars)
    chunk_store = algo.reset_chunk_store() if iterative else {}

    k = (chunki_states, chunki_points)
    cstore = {k: chunk_store[k]} if k in chunk_store else {}
    return results, cstore


class LocalClusterEngine(DaskBaseEngine):
    """
    The dask engine for foxes calculations on a local cluster.

    Attributes
    ----------
    cluster_pars: dict
        Parameters for the cluster
    client_pars: dict
        Parameters for the client of the cluster

    :group: engines

    """

    def __init__(
        self,
        *args,
        cluster_pars={},
        client_pars={},
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        args: tuple, optional
            Additional parameters for the DaskBaseEngine class
        cluster_pars: dict
            Parameters for the cluster
        client_pars: dict
            Parameters for the client of the cluster
        kwargs: dict, optional
            Additional parameters for the base class

        """
        super().__init__(*args, **kwargs)

        load_distributed()

        self.cluster_pars = cluster_pars
        self.client_pars = client_pars

        self.dask_config["scheduler"] = "distributed"
        self.dask_config["distributed.scheduler.worker-ttl"] = None

        self._cluster = None
        self._client = None

    def __enter__(self):
        self.print("Launching local dask cluster..")
        self._cluster = distributed.LocalCluster(
            n_workers=self.n_workers, **self.cluster_pars
        ).__enter__()
        self._client = distributed.Client(self._cluster, **self.client_pars).__enter__()
        self.print(self._cluster)
        self.print(f"Dashboard: {self._client.dashboard_link}\n")
        return super().__enter__()

    def __exit__(self, *args):
        self.print(f"Shutting down {type(self._cluster).__name__}")
        # self._client.retire_workers()
        # from time import sleep
        # sleep(1)
        # self._client.shutdown()
        self._client.__exit__(*args)
        self._cluster.__exit__(*args)
        super().__exit__(*args)

    def __del__(self):
        if hasattr(self, "_client") and self._client is not None:
            self._client.__del__()
        if hasattr(self, "_cluster") and self._cluster is not None:
            self._cluster.__del__()
        super().__del__()

    def submit(self, f, *args, **kwargs):
        """
        Submits a job to worker, obtaining a future

        Parameters
        ----------
        f: Callable
            The function f(*args, **kwargs) to be
            submitted
        args: tuple, optional
            Arguments for the function
        kwargs: dict, optional
            Arguments for the function

        Returns
        -------
        future: object
            The future object

        """
        return self._client.submit(f, *args, **kwargs)

    def await_result(self, future):
        """
        Waits for result from a future

        Parameters
        ----------
        future: object
            The future

        Returns
        -------
        result: object
            The calculation result

        """
        return future.result()

    def run_calculation(
        self,
        algo,
        model,
        model_data=None,
        farm_data=None,
        point_data=None,
        out_vars=[],
        chunk_store={},
        sel=None,
        isel=None,
        iterative=False,
        **calc_pars,
    ):
        """
        Runs the model calculation

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The algorithm object
        model: foxes.core.DataCalcModel
            The model that whose calculate function
            should be run
        model_data: xarray.Dataset
            The initial model data
        farm_data: xarray.Dataset
            The initial farm data
        point_data: xarray.Dataset
            The initial point data
        out_vars: list of str, optional
            Names of the output variables
        chunk_store: foxes.utils.Dict
            The chunk store
        sel: dict, optional
            Selection of coordinate subsets
        isel: dict, optional
            Selection of coordinate subsets index values
        iterative: bool
            Flag for use within the iterative algorithm
        calc_pars: dict, optional
            Additional parameters for the model.calculate()

        Returns
        -------
        results: xarray.Dataset
            The model results

        """
        # subset selection:
        model_data, farm_data, point_data = self.select_subsets(
            model_data, farm_data, point_data, sel=sel, isel=isel
        )

        # basic checks:
        super().run_calculation(algo, model, model_data, farm_data, point_data)

        # prepare:
        n_states = model_data.sizes[FC.STATE]
        out_coords = model.output_coords()
        coords = {}
        if FC.STATE in out_coords and FC.STATE in model_data.coords:
            coords[FC.STATE] = model_data[FC.STATE].to_numpy()
        if farm_data is None:
            farm_data = xr.Dataset()
        goal_data = farm_data if point_data is None else point_data

        # calculate chunk sizes:
        n_targets = point_data.sizes[FC.TARGET] if point_data is not None else 0
        chunk_sizes_states, chunk_sizes_targets = self.calc_chunk_sizes(
            n_states, n_targets
        )
        n_chunks_states = len(chunk_sizes_states)
        n_chunks_targets = len(chunk_sizes_targets)
        self.print(
            f"Selecting n_chunks_states = {n_chunks_states}, n_chunks_targets = {n_chunks_targets}",
            level=2,
        )

        # scatter algo and model:
        n_chunks_all = n_chunks_states * n_chunks_targets
        falgo = self._client.scatter(algo, broadcast=True)
        fmodel = self._client.scatter(model, broadcast=True)
        cpars = self._client.scatter(calc_pars, broadcast=True)
        all_data = [falgo, fmodel, cpars]

        # submit chunks:
        self.print(f"Submitting {n_chunks_all} chunks to {self.n_workers} workers")
        pbar = (
            tqdm(total=n_chunks_all)
            if self.verbosity > 1 and self.has_progress_bar
            else None
        )
        futures = {}
        i0_states = 0
        pdone = -1
        counter = 0
        for chunki_states in range(n_chunks_states):
            i1_states = i0_states + chunk_sizes_states[chunki_states]
            i0_targets = 0
            for chunki_points in range(n_chunks_targets):
                i1_targets = i0_targets + chunk_sizes_targets[chunki_points]

                # get this chunk's data:
                data = self.get_chunk_input_data(
                    algo=algo,
                    model_data=model_data,
                    farm_data=farm_data,
                    point_data=point_data,
                    states_i0_i1=(i0_states, i1_states),
                    targets_i0_i1=(i0_targets, i1_targets),
                    out_vars=out_vars,
                    chunki_states=chunki_states,
                    chunki_points=chunki_points,
                    n_chunks_states=n_chunks_states,
                    n_chunks_points=n_chunks_targets,
                )

                # scatter data:
                fut_data = []
                names = []
                dims = []
                ldims = [d.loop_dims for d in data]
                for dt in data:
                    for k, d in dt.items():
                        fut_data.append(self._client.scatter(d, hash=False))
                        names.append(k)
                        dims.append(dt.dims[k])
                names = self._client.scatter(names)
                dims = self._client.scatter(dims)
                ldims = self._client.scatter(ldims)
                all_data += [fut_data, names, dims, ldims]

                # scatter chunk store data:
                cstore = chunk_store
                if len(cstore):
                    cstore = self._client.scatter(cstore, hash=False)
                    all_data.append(cstore)

                # submit model calculation:
                futures[(chunki_states, chunki_points)] = self.submit(
                    _run_on_cluster,
                    falgo,
                    fmodel,
                    *fut_data,
                    names=names,
                    dims=dims,
                    mdata_size=len(data[0]),
                    fdata_size=len(data[1]),
                    iterative=iterative,
                    chunk_store=cstore,
                    chunki_states=chunki_states,
                    chunki_points=chunki_points,
                    n_chunks_states=n_chunks_states,
                    n_chunks_points=n_chunks_targets,
                    i0_states=i0_states,
                    cpars=cpars,
                    retries=10,
                )
                del fut_data, cstore

                i0_targets = i1_targets

                counter += 1
                if pbar is not None:
                    pbar.update()
                elif self.verbosity > 1 and self.prints_progress:
                    pr = int(100 * counter / n_chunks_all)
                    if pr > pdone:
                        pdone = pr
                        print(
                            f"{self.name}: Submitted {counter} of {n_chunks_all} chunks, {pdone}%"
                        )
            i0_states = i1_states

        del falgo, fmodel, farm_data, point_data, calc_pars
        if pbar is not None:
            pbar.close()
        elif self.verbosity > 1 and self.prints_progress:
            print(f"{type(self).__name__}: Submitted all {counter} chunks\n")

        results = self.combine_results(
            algo=algo,
            futures=futures,
            model_data=model_data,
            out_vars=out_vars,
            out_coords=out_coords,
            n_chunks_states=n_chunks_states,
            n_chunks_targets=n_chunks_targets,
            goal_data=goal_data,
            iterative=iterative,
        ).persist()

        # self._client.cancel(all_data)

        return results


class SlurmClusterEngine(LocalClusterEngine):
    """
    The dask engine for foxes calculations on a SLURM cluster.

    :group: engines

    """

    def __enter__(self):
        self.print("Launching dask cluster on HPC using SLURM..")
        cargs = deepcopy(self.cluster_pars)
        nodes = cargs.pop("nodes", 1)

        dask_jobqueue = import_module(
            "dask_jobqueue",
            pip_hint="pip install setuptools dask-jobqueue",
            conda_hint="conda install setuptools dask-jobqueue -c conda-forge",
        )

        self._cluster = dask_jobqueue.SLURMCluster(**cargs)
        self._cluster.scale(futures=nodes)
        self._cluster = self._cluster.__enter__()
        self._client = distributed.Client(self._cluster, **self.client_pars).__enter__()

        self.print(self._cluster)
        self.print(f"Dashboard: {self._client.dashboard_link}\n")
        print(self._cluster.job_script())

        return DaskBaseEngine.__enter__(self)



from .iterative import Iterative as Iterative

from .models.convergence import ConvCrit as ConvCrit
from .models.convergence import ConvCritList as ConvCritList
from .models.convergence import ConvVarDelta as ConvVarDelta
from .models.convergence import DefaultConv as DefaultConv

from . import models as models



import numpy as np
import time
import argparse
from pathlib import Path

import flappy as fl
from flappy.config.variables import variables as FV

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-o", "--ofile", help="The output file name", default="results.csv.gz"
    )
    parser.add_argument(
        "--n_cpus", help="The number of processors", type=int, default=1
    )
    args = parser.parse_args()

    n_s = 99
    n_t = 84
    wd = 88.1
    ti = 0.04
    rotor = "centre"
    c = 100
    p0 = np.array([0.0, 0.0])
    stp = np.array([533.0, 12.0])
    tfile = "../NREL-5MW-D126-H90.csv"
    ofile = Path(args.ofile)

    # init flappy:
    fl.init_flappy(n_cpus=args.n_cpus)

    # load model book:
    mbook = fl.ModelBook(ct_power_curve_file=tfile)

    # create wind farm:
    farm = fl.WindFarm()
    fl.input.add_turbine_row(
        farm,
        rotor_diameter=126.0,
        hub_height=90.0,
        rotor_model=rotor,
        wake_models=["Bastankhah_rotor"],
        turbine_models=["ct_P_curves"],
        base_point=p0,
        step_vector=stp,
        steps=n_t - 1,
    )

    # create states:
    ws0 = 6.0
    ws1 = 16.0
    states = fl.input.AFSScan(
        ws_min=ws0,
        ws_delta=(ws1 - ws0) / (n_s - 1),
        ws_n_bins=n_s,
        func_pdf_ws=None,
        wd_min=wd,
        wd_delta=1.0,
        wd_n_bins=1,
        func_pdf_wd=None,
        ti_min=ti,
        ti_delta=0.01,
        ti_n_bins=1,
        func_pdf_ti=None,
        rho_min=1.225,
        rho_delta=0.001,
        rho_n_bins=1,
        func_pdf_rho=None,
        max_chunk_size=c,
    )
    states.initialize()

    time0 = time.time()

    # run calculation:
    results = farm.calculate(mbook, states, wake_superp=["wind_linear"])

    time1 = time.time()
    print("\nCalc time =", time1 - time0, "\n")

    df = results.state_turbine_results[[FV.WD, FV.AMB_WS, FV.WS, FV.AMB_P, FV.P]]

    print()
    print("TRESULTS\n")
    print(df)

    print("\nWriting file", ofile)
    df.to_csv(ofile)

    # close flappy:
    fl.shutdown_flappy()



"""
Functions for adding turbines to the wind farm.
"""

from .row import add_row as add_row
from .grid import add_grid as add_grid
from .from_json import add_from_json as add_from_json
from .from_csv import add_from_csv as add_from_csv
from .from_eww import add_from_eww as add_from_eww
from .from_wrf import add_from_wrf as add_from_wrf
from .from_file import add_from_file as add_from_file
from .from_df import add_from_df as add_from_df
from .from_random import add_random as add_random
from .ring import add_ring as add_ring
from .from_arrays import add_from_arrays as add_from_arrays



import argparse
import matplotlib.pyplot as plt
import numpy as np
from pathlib import Path

import foxes
import foxes.variables as FV

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-a", "--animation", help="Write flow animation file", action="store_true"
    )
    parser.add_argument(
        "-A",
        "--ani_file",
        help="Path to the animation file to be written",
        default="ani.gif",
    )
    parser.add_argument(
        "-F",
        "--fps",
        help="The frames per second value for the animation",
        type=int,
        default=4,
    )
    parser.add_argument("--debug", help="Switch on wake debugging", action="store_true")
    parser.add_argument(
        "-S",
        "--n_states",
        help="The number of states",
        type=int,
        default=50,
    )
    parser.add_argument(
        "-l",
        "--layout",
        help="The wind farm layout file (path or static)",
        default="test_farm_67.csv",
    )
    parser.add_argument(
        "-t",
        "--turbine_file",
        help="The P-ct-curve csv file (path or static)",
        default="NREL-5MW-D126-H90.csv",
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-p", "--pwakes", help="The partial wakes models", default=None, nargs="+"
    )
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["Bastankhah2014_linear_lim_k004"],
        nargs="+",
    )
    parser.add_argument(
        "-f", "--frame", help="The wake frame", default="seq_dyn_wakes_1min"
    )
    parser.add_argument(
        "-m", "--tmodels", help="The turbine models", default=[], nargs="+"
    )
    parser.add_argument(
        "-sl",
        "--show_layout",
        help="Flag for showing layout figure",
        action="store_true",
    )
    parser.add_argument(
        "-V",
        "--variable",
        help="Plot variable",
        default=FV.WS,
    )
    parser.add_argument(
        "-VMAX",
        "--max_variable",
        help="Maximal plot variable",
        default=10,
        type=float,
    )
    parser.add_argument("-e", "--engine", help="The engine", default="NumpyEngine")
    parser.add_argument(
        "-d", "--deflection", help="The wake deflection", default="no_deflection"
    )
    parser.add_argument(
        "-y",
        "--yawm",
        help="The uniform yaw misalignment value",
        type=float,
        default=None,
    )
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-nf", "--nofig", help="Do not show figures", action="store_true"
    )
    args = parser.parse_args()

    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(args.turbine_file)
    mbook.turbine_types[ttype.name] = ttype

    # optionally set turbines in yaw:
    N = 3
    if args.yawm is None:
        ymodels = []
    else:
        yawm = np.zeros((args.n_states, N * N), dtype=np.float64)
        yawm[:, :N] = args.yawm
        mbook.turbine_models["set_yawm"] = foxes.models.turbine_models.SetFarmVars(
            pre_rotor=True
        )
        mbook.turbine_models["set_yawm"].add_var(FV.YAWM, yawm)
        ymodels = ["set_yawm"]

    states = foxes.input.states.Timeseries(
        data_source="timeseries_3000.csv.gz",
        output_vars=[FV.WS, FV.WD, FV.TI, FV.RHO],
        # fixed_vars={FV.WD: 270},
        var2col={FV.WD: "WD", FV.WS: "WS", FV.TI: "TI", FV.RHO: "RHO"},
        states_sel=range(240, 240 + args.n_states),
    )

    farm = foxes.WindFarm()
    foxes.input.farm_layout.add_grid(
        farm,
        xy_base=np.array([0.0, 0.0]),
        step_vectors=np.array([[1000.0, 0], [0, 800.0]]),
        steps=(N, N),
        turbine_models=ymodels + args.tmodels + [ttype.name],
    )

    if not args.nofig and args.show_layout:
        ax = foxes.output.FarmLayoutOutput(farm).get_figure()
        plt.show()
        plt.close(ax.get_figure(figsize=(8, 8)))

    with foxes.Engine.new(
        engine_type=args.engine,
        n_procs=args.n_cpus,
        chunk_size_states=args.chunksize_states,
        chunk_size_points=args.chunksize_points,
    ):
        algo = foxes.algorithms.Sequential(
            farm,
            states,
            rotor_model=args.rotor,
            wake_models=args.wakes,
            wake_frame=args.frame,
            wake_deflection=args.deflection,
            partial_wakes=args.pwakes,
            mbook=mbook,
            verbosity=0,
        )

        # in case of animation, add a plugin that creates the images:
        if args.animation:
            anigen = foxes.output.SeqFlowAnimationPlugin(
                orientation="xy",
                var=args.variable,
                resolution=10,
                levels=None,
                quiver_pars=dict(scale=0.01),
                quiver_n=307,
                xmin=-1000,
                ymin=-1000,
                xmax=4000,
                ymax=3000,
                vmin=0,
                vmax=args.max_variable,
                title=None,
                animated=True,
                rotor_color="red",
            )
            algo.plugins.append(anigen)

            if args.debug:
                anigen_debug = foxes.output.SeqWakeDebugPlugin()
                algo.plugins.append(anigen_debug)

        # run all states sequentially:
        for r in algo:
            print(algo.index)

    print("\nFarm results:\n")
    print(algo.farm_results)

    if args.animation:
        print("\nCalculating animation")

        fig, ax = plt.subplots()
        anim = foxes.output.Animator(fig)
        anim.add_generator(anigen.gen_images(ax))
        if args.debug:
            anim.add_generator(anigen_debug.gen_images(ax))
        ani = anim.animate(interval=600)

        lo = foxes.output.FarmLayoutOutput(farm)
        lo.get_figure(
            fig=fig,
            ax=ax,
            title="",
            annotate=1,
            anno_delx=-120,
            anno_dely=-60,
            alpha=0,
            s=10,
        )

        fpath = Path(args.ani_file)
        print("Writing file", fpath)
        if fpath.suffix == ".gif":
            ani.save(filename=fpath, writer="pillow", fps=args.fps)
        else:
            ani.save(filename=fpath, writer="ffmpeg", fps=args.fps)



"""
Atmospheric input states.
"""

from .single import SingleStateStates as SingleStateStates
from .scan import ScanStates as ScanStates
from .wrg_states import WRGStates as WRGStates
from .weibull_sectors import WeibullSectors as WeibullSectors
from .dataset_states import DatasetStates as DatasetStates

from .states_table import StatesTable as StatesTable
from .states_table import Timeseries as Timeseries
from .states_table import TabStates as TabStates

from .field_data import FieldData as FieldData
from .field_data import WeibullField as WeibullField

from .point_cloud_data import WeibullPointCloud as WeibullPointCloud
from .point_cloud_data import PointCloudData as PointCloudData

from .multi_height import MultiHeightStates as MultiHeightStates
from .multi_height import MultiHeightTimeseries as MultiHeightTimeseries
from .multi_height import MultiHeightNCStates as MultiHeightNCStates
from .multi_height import MultiHeightNCTimeseries as MultiHeightNCTimeseries

from .one_point_flow import OnePointFlowStates as OnePointFlowStates
from .one_point_flow import OnePointFlowTimeseries as OnePointFlowTimeseries
from .one_point_flow import (
    OnePointFlowMultiHeightTimeseries as OnePointFlowMultiHeightTimeseries,
)
from .one_point_flow import (
    OnePointFlowMultiHeightNCTimeseries as OnePointFlowMultiHeightNCTimeseries,
)

from .newa_states import NEWAStates as NEWAStates

from . import create as create



from abc import abstractmethod

from foxes.utils import new_instance
from .model import Model


class WakeDeflection(Model):
    """
    Abstract base class for wake deflection models.

    :group: core

    """

    @property
    def has_uv(self):
        """
        This model uses wind vector data

        Returns
        -------
        hasuv: bool
            Flag for wind vector data

        """
        return False

    @abstractmethod
    def calc_deflection(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        coos,
    ):
        """
        Calculates the wake deflection.

        This function optionally adds FC.WDEFL_ROT_ANGLE or
        FC.WDEFL_DWS_FACTOR to the tdata.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order
        coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)

        Returns
        -------
        coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)

        """
        pass

    def get_yaw_alpha_seq(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        x,
    ):
        """
        Computes sequential wind vector rotation angles.

        Wind vector rotation angles are computed at the
        current trace points due to a yawed rotor
        for sequential runs.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order
        x: numpy.ndarray
            The distance from the wake causing rotor
            for the first n_times subsequent time steps,
            shape: (n_times,)

        Returns
        -------
        alpha: numpy.ndarray
            The delta WD result at the x locations,
            shape: (n_times,)

        """
        raise NotImplementedError(
            f"Wake deflection '{self.name}' not implemented for sequential runs"
        )

    @classmethod
    def new(cls, wdefl_type, *args, **kwargs):
        """
        Run-time wake deflection model factory.

        Parameters
        ----------
        wdefl_type: str
            The selected derived class name
        args: tuple, optional
            Additional parameters for constructor
        kwargs: dict, optional
            Additional parameters for constructor

        """
        return new_instance(cls, wdefl_type, *args, **kwargs)



import numpy as np

import foxes
import foxes.variables as FV
import foxes.constants as FC


def test_set_farm_vars():
    n_s = 360
    n_tr = 3

    wd = np.arange(0.0, 360.0, 360 / n_s)
    states = foxes.input.states.ScanStates(
        scans={
            FV.WD: wd,
            FV.WS: [9.0],
            FV.TI: [0.04],
            FV.RHO: [1.225],
        },
    )

    n_t = n_tr**2
    x = np.zeros((n_s, n_t), dtype=foxes.config.dtype_double)
    x[:] = wd[:, None] + np.arange(n_t)[None, :] / 10

    farm = foxes.WindFarm()
    foxes.input.farm_layout.add_grid(
        farm,
        xy_base=[0.0, 0.0],
        step_vectors=[[800.0, 0.0], [0.0, 800.0]],
        steps=[n_tr, n_tr],
        turbine_models=["NREL5MW", "set_x"],
        verbosity=0,
    )

    with foxes.Engine.new("default", verbosity=0):
        for pr in [False, True]:
            print(f"\npre_rotor = {pr}\n")

            mbook = foxes.ModelBook()
            mbook.turbine_models["set_x"] = foxes.models.turbine_models.SetFarmVars(
                pre_rotor=pr
            )
            mbook.turbine_models["set_x"].add_var("x", x)

            algo = foxes.algorithms.Downwind(
                farm=farm,
                states=states,
                wake_models=["Bastankhah2014_linear_lim_k004"],
                mbook=mbook,
                verbosity=0,
            )

            farm_results = algo.calc_farm()

            fr = farm_results.to_dataframe()
            print(fr[[FV.WD, "x"]])

            for i, g in fr.reset_index().groupby(FC.TURBINE):
                assert np.allclose(g["x"].values, g[FV.WD].values + i / 10)


if __name__ == "__main__":
    test_set_farm_vars()



from abc import abstractmethod
import numpy as np

from foxes.core import SingleTurbineWakeModel
from foxes.config import config
import foxes.variables as FV


class DistSlicedWakeModel(SingleTurbineWakeModel):
    """
    Abstract base class for wake models for which
    the x-denpendency can be separated from the
    yz-dependency.

    The multi-yz ability is used by the `PartialDistSlicedWake`
    partial wakes model.

    :group: models.wake_models

    """

    def new_wake_deltas(self, algo, mdata, fdata, tdata):
        """
        Creates new empty wake delta arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        wake_deltas: dict
            Key: variable name, value: The zero filled
            wake deltas, shape: (n_states, n_targets, n_tpoints, ...)

        """
        if self.has_uv:
            duv = np.zeros(
                (tdata.n_states, tdata.n_targets, tdata.n_tpoints, 2),
                dtype=config.dtype_double,
            )
            return {FV.UV: duv}
        else:
            dws = np.zeros(
                (tdata.n_states, tdata.n_targets, tdata.n_tpoints),
                dtype=config.dtype_double,
            )
            return {FV.WS: dws}

    @abstractmethod
    def calc_wakes_x_yz(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        x,
        yz,
    ):
        """
        Calculate wake deltas.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The x values, shape: (n_states, n_targets)
        yz: numpy.ndarray
            The yz values for each x value, shape:
            (n_states, n_targets, n_yz_per_target, 2)

        Returns
        -------
        wdeltas: dict
            The wake deltas. Key: variable name str,
            value: numpy.ndarray, shape: (n_st_sel, n_yz_per_target)
        st_sel: numpy.ndarray of bool
            The state-target selection, for which the wake
            is non-zero, shape: (n_states, n_targets)

        """
        pass

    def contribute(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        wake_coos,
        wake_deltas,
    ):
        """
        Modifies wake deltas at target points by
        contributions from the specified wake source turbines.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints, ...)

        """
        x = wake_coos[:, :, 0, 0]
        yz = wake_coos[..., 1:3]

        wdeltas, st_sel = self.calc_wakes_x_yz(
            algo, mdata, fdata, tdata, downwind_index, x, yz
        )

        if self.affects_ws and self.has_uv:
            assert self.has_vector_wind_superp, (
                f"Wake model {self.name}: Missing vector wind superposition, got '{self.wind_superposition}'"
            )
            if FV.UV in wdeltas or FV.WS in wdeltas:
                if FV.UV not in wdeltas:
                    self.vec_superp.wdeltas_ws2uv(
                        algo, fdata, tdata, downwind_index, wdeltas, st_sel
                    )
                wake_deltas[FV.UV] = self.vec_superp.add_wake_vector(
                    algo,
                    mdata,
                    fdata,
                    tdata,
                    downwind_index,
                    st_sel,
                    wake_deltas[FV.UV],
                    wdeltas.pop(FV.UV),
                )

        for v, hdel in wdeltas.items():
            try:
                superp = self.superp[v]
            except KeyError:
                raise KeyError(
                    f"Model '{self.name}': Missing wake superposition entry for variable '{v}', found {sorted(list(self.superp.keys()))}"
                )

            wake_deltas[v] = superp.add_wake(
                algo,
                mdata,
                fdata,
                tdata,
                downwind_index,
                st_sel,
                v,
                wake_deltas[v],
                hdel,
            )



import time
import argparse
import matplotlib.pyplot as plt

import foxes
import foxes.variables as FV

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-l",
        "--layout",
        help="The wind farm layout file (path or static)",
        default="test_farm_67.csv",
    )
    parser.add_argument(
        "-s",
        "--states",
        help="The timeseries input file (path or static)",
        default="timeseries_8000.csv.gz",
    )
    parser.add_argument(
        "-t",
        "--turbine_file",
        help="The P-ct-curve csv file (path or static)",
        default="NREL-5MW-D126-H90.csv",
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-p", "--pwakes", help="The partial wakes models", default=None, nargs="+"
    )
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["Jensen_linear_k007"],
        nargs="+",
    )
    parser.add_argument("-f", "--frame", help="The wake frame", default="rotor_wd")
    parser.add_argument(
        "-m", "--tmodels", help="The turbine models", default=[], nargs="+"
    )
    parser.add_argument(
        "-sl",
        "--show_layout",
        help="Flag for showing layout figure",
        action="store_true",
    )
    parser.add_argument(
        "-ts",
        "--timestep",
        help="The timestep of the input timeseries or data in minutes",
        default=60 * 24 * 365,  # default is one year
    )
    parser.add_argument(
        "-it", "--iterative", help="Use iterative algorithm", action="store_true"
    )
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=5000,
        type=int,
    )
    parser.add_argument(
        "-nf", "--nofig", help="Do not show figures", action="store_true"
    )
    args = parser.parse_args()

    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(args.turbine_file)
    mbook.turbine_types[ttype.name] = ttype

    states = foxes.input.states.Timeseries(
        data_source=args.states,
        output_vars=[FV.WS, FV.WD, FV.TI, FV.RHO],
        var2col={FV.WS: "ws", FV.WD: "wd", FV.TI: "ti"},
        fixed_vars={FV.RHO: 1.225},
    )

    farm = foxes.WindFarm()
    foxes.input.farm_layout.add_from_file(
        farm, args.layout, turbine_models=args.tmodels + [ttype.name]
    )

    if not args.nofig and args.show_layout:
        ax = foxes.output.FarmLayoutOutput(farm).get_figure()
        plt.show()
        plt.close(ax.get_figure())

    Algo = foxes.algorithms.Iterative if args.iterative else foxes.algorithms.Downwind
    algo = Algo(
        farm,
        states=states,
        rotor_model=args.rotor,
        wake_models=args.wakes,
        wake_frame=args.frame,
        partial_wakes=args.pwakes,
        mbook=mbook,
    )

    cluster_pars = {
        "nodes": 1,  # number of nodes
        "cores": 24,  # number of cores per node
        "processes": 4,  # number of workers per node
        "memory": "64GB",  # memory per node
        "walltime": "00:00:10",
        "queue": "cfds.p",
        # "silence_logs": "info", # print all info log
    }

    with foxes.Engine.new(
        engine_type="slurm_cluster",
        n_procs=4,
        cluster_pars=cluster_pars,
        chunk_size_states=args.chunksize_states,
        chunk_size_points=args.chunksize_points,
    ):
        time0 = time.time()
        farm_results = algo.calc_farm()
        time1 = time.time()

        print("\nCalc time =", time1 - time0, "\n")

        o = foxes.output.FarmResultsEval(farm_results)
        o.add_capacity(algo)
        o.add_capacity(algo, ambient=True)
        o.add_efficiency()

        print("\nFarm results:\n")
        print(farm_results)

        # state-turbine results
        farm_df = farm_results.to_dataframe()
        print("\nFarm results data:\n")
        print(
            farm_df[
                [
                    FV.X,
                    FV.WD,
                    FV.AMB_REWS,
                    FV.REWS,
                    FV.AMB_TI,
                    FV.TI,
                    FV.AMB_P,
                    FV.P,
                    FV.EFF,
                ]
            ]
        )

        # results by turbine
        turbine_results = o.reduce_states(
            {
                FV.AMB_P: "weights",
                FV.P: "weights",
                FV.AMB_CAP: "weights",
                FV.CAP: "weights",
            }
        )
        turbine_results[FV.AMB_YLD] = o.calc_turbine_yield(
            algo=algo, annual=True, ambient=True
        )
        turbine_results[FV.YLD] = o.calc_turbine_yield(algo=algo, annual=True)
        turbine_results[FV.EFF] = turbine_results[FV.P] / turbine_results[FV.AMB_P]
        print("\nResults by turbine:\n")
        print(turbine_results)

        # power results
        P0 = o.calc_mean_farm_power(ambient=True)
        P = o.calc_mean_farm_power()
        print(f"\nFarm power        : {P / 1000:.1f} MW")
        print(f"Farm ambient power: {P0 / 1000:.1f} MW")
        print(f"Farm efficiency   : {o.calc_farm_efficiency() * 100:.2f} %")
        print(f"Annual farm yield : {turbine_results[FV.YLD].sum():.2f} GWh")
